import argparse
from collections import defaultdict
import re
from copy import deepcopy, copy

from sortedcontainers import SortedList
import statistics

from rck.core.io import EXTERNAL_NA_ID
from rck.core.structures import Strand, Adjacency, Position
from rck.utils.adj.convert import GUNDEM_PER_SAMPLE_SUPPORT

ORIGIN_IDS = "origin_ids"

KEEP = "keep"
REMOVE = "remove"


class Merger(object):
    def __init__(self, origin_ids_field=ORIGIN_IDS):
        self.origin_ids_field = origin_ids_field
        self.adjs_by_ids = {}
        self.merged_position_pairs = set()
        self.ajds_ids_by_positions = defaultdict(set)
        self.data = defaultdict(lambda: {
            Strand.FORWARD: SortedList(),
            Strand.REVERSE: SortedList(),
        })

    def add_adjacency(self, adjacency, force_sv_type=False, force_na_type=False, max_distance=-1):
        aid = adjacency.extra.get(EXTERNAL_NA_ID, adjacency.idx)
        lp, rp = adjacency.position1, adjacency.position2
        chrl, chrr = lp.chromosome, rp.chromosome
        coordl, coordr = lp.coordinate, rp.coordinate
        strandl, strandr = lp.strand, rp.strand
        lp_neighbours = self.get_closest_coordinates(chromosome=chrl, coordinate=coordl, strand=strandl, max_distance=max_distance)
        rp_neighbours = self.get_closest_coordinates(chromosome=chrr, coordinate=coordr, strand=strandr, max_distance=max_distance)
        if lp_neighbours is None or rp_neighbours is None:
            self.__insert_adjacency(adjacency=adjacency)
            self.adjs_by_ids[aid] = adjacency
            self.add_adjs_id_by_position(aid=aid, chromosome=chrl, strand=strandl, coordinate=coordl)
            self.add_adjs_id_by_position(aid=aid, chromosome=chrr, strand=strandr, coordinate=coordr)
            self.merged_position_pairs.add(((chrl, strandl, coordl), (chrr, strandr, coordr)))
            return
        neighbour_pairs = []
        for ln in lp_neighbours:
            for rn in rp_neighbours:
                if chrl == chrr and strandl == strandr and ln >= rn:
                    continue
                neighbour_pairs.append((ln, rn))
        if len(neighbour_pairs) == 0:
            neighbour_pairs = [(-1, -1)]
            fall_through = True
        else:
            fall_through = False
            neighbour_pairs = sorted(neighbour_pairs, key=lambda e: abs(e[0] - coordl) + abs(e[1] - coordr))
        for lpn, rpn in neighbour_pairs:
            if fall_through:
                break
            lpn_adj_ids = self.get_adjs_ids_by_position(chromosome=chrl, strand=strandl, coordinate=lpn)
            rpn_adj_ids = self.get_adjs_ids_by_position(chromosome=chrr, strand=strandr, coordinate=rpn)
            common_adjs_ids = lpn_adj_ids & rpn_adj_ids
            if len(common_adjs_ids) == 0:
                continue
            # here we merge the stuffffff
            common_adjs = [self.adjs_by_ids[common_aid] for common_aid in common_adjs_ids]
            lp_coordinates = []
            rp_coordinates = []
            aid_flipped = {}
            for adj in common_adjs:
                p1_chr, p2_chr = adj.position1.chromosome, adj.position2.chromosome
                p1_coord, p2_coord = adj.position1.coordinate, adj.position2.coordinate
                p1_strand, p2_strand = adj.position1.strand, adj.position2.strand
                if (p1_chr, p1_strand) == (chrl, strandl) and (p2_chr, p2_strand) == (chrr, strandr):
                    aid_flipped[adj.extra.get(EXTERNAL_NA_ID, adj.idx)] = False
                    lp_coordinates.append(p1_coord)
                    rp_coordinates.append(p2_coord)
                elif (p1_chr, p1_strand) == (chrr, strandr) and (p2_chr, p2_strand) == (chrl, strandl):
                    aid_flipped[adj.extra.get(EXTERNAL_NA_ID, adj.idx)] = True
                    lp_coordinates.append(p2_coord)
                    rp_coordinates.append(p1_coord)
                else:
                    raise Exception("Problems in merging")
            lp_coordinates.append(coordl)
            rp_coordinates.append(coordr)
            new_lp_coordinate = int(statistics.mean(lp_coordinates))
            new_rp_coordinate = int(statistics.mean(rp_coordinates))
            if common_adjs_ids == lpn_adj_ids:
                self.data[chrl][strandl].discard(lpn)
            if common_adjs_ids == rpn_adj_ids:
                self.data[chrr][strandr].discard(rpn)
            if new_lp_coordinate not in self.data[chrl][strandl]:
                self.data[chrl][strandl].add(new_lp_coordinate)
            if new_rp_coordinate not in self.data[chrr][strandr]:
                self.data[chrr][strandr].add(new_rp_coordinate)
            assert ((chrl, strandl, lpn), (chrr, strandr, rpn)) in self.merged_position_pairs or ((chrr, strandr, rpn), (chrl, strandl, lpn)) in self.merged_position_pairs
            self.merged_position_pairs.discard(((chrl, strandl, lpn), (chrr, strandr, rpn)))
            self.merged_position_pairs.discard(((chrr, strandr, rpn), (chrl, strandl, lpn)))
            self.merged_position_pairs.add(((chrl, strandl, new_lp_coordinate), (chrr, strandr, new_rp_coordinate)))
            for common_adj_id in common_adjs_ids:
                # l_coord, r_coord = (rpn, lpn) if aid_flipped[common_adj_id] else (lpn, rpn)
                self.delete_adjs_id_by_position(aid=common_adj_id, chromosome=chrl, coordinate=lpn, strand=strandl)
                self.delete_adjs_id_by_position(aid=common_adj_id, chromosome=chrr, coordinate=rpn, strand=strandr)
            self.adjs_by_ids[aid] = adjacency
            for current_aid in sorted(common_adjs_ids) + [aid]:
                self.add_adjs_id_by_position(aid=current_aid, chromosome=chrl, coordinate=new_lp_coordinate, strand=strandl)
                self.add_adjs_id_by_position(aid=current_aid, chromosome=chrr, coordinate=new_rp_coordinate, strand=strandr)
            break
        else:
            self.__insert_adjacency(adjacency=adjacency)
            self.adjs_by_ids[adjacency.extra.get(EXTERNAL_NA_ID, adjacency.idx)] = adjacency
            self.add_adjs_id_by_position(aid=aid, chromosome=chrl, strand=strandl, coordinate=coordl)
            self.add_adjs_id_by_position(aid=aid, chromosome=chrr, strand=strandr, coordinate=coordr)
            self.merged_position_pairs.add(((chrl, strandl, coordl), (chrr, strandr, coordr)))

    def __insert_adjacency(self, adjacency):
        if adjacency.position1.coordinate not in self.data[adjacency.position1.chromosome][adjacency.position1.strand]:
            self.data[adjacency.position1.chromosome][adjacency.position1.strand].add(adjacency.position1.coordinate)
        if adjacency.position2.coordinate not in self.data[adjacency.position2.chromosome][adjacency.position2.strand]:
            self.data[adjacency.position2.chromosome][adjacency.position2.strand].add(adjacency.position2.coordinate)

    def get_closest_coordinates(self, chromosome, coordinate, strand, max_distance=-1):
        coordinates = self.data[chromosome][strand]
        if len(coordinates) == 0:
            return None
        if coordinate in coordinates:
            return [coordinate]
        left_index = coordinates.bisect_left(coordinate)
        neighbours = []
        for i in reversed(range(0, left_index)):
            other_coordinate = coordinates[i]
            if abs(other_coordinate - coordinate) < max_distance:
                neighbours.append(other_coordinate)
            else:
                break
        for i in range(left_index, len(coordinates)):
            other_coordinate = coordinates[i]
            if abs(other_coordinate - coordinate) < max_distance:
                neighbours.append(other_coordinate)
            else:
                break
        if len(neighbours) == 0:
            return None
        return sorted(neighbours, key=lambda e: abs(e - coordinate))

    def get_closest_coordinate_right(self, chromosome, coordinate, strand, max_distance=1):
        coordinates = self.data[chromosome][strand]
        if coordinate in coordinates:
            return coordinate
        right_coordinate = coordinates.bisect_right(coordinate)
        if len(coordinates) == 0 or right_coordinate == len(coordinates):
            return None
        right_neighbour = coordinates[right_coordinate]
        if max_distance >= 0 and abs(right_neighbour - coordinate) > max_distance:
            return None
        return right_neighbour

    def get_adjs_ids_by_position(self, chromosome, coordinate, strand):
        return self.ajds_ids_by_positions.get((chromosome, strand, coordinate), set())

    def add_adjs_id_by_position(self, aid, chromosome, coordinate, strand):
        self.ajds_ids_by_positions[(chromosome, strand, coordinate)].add(aid)

    def delete_adjs_id_by_position(self, aid, chromosome, coordinate, strand):
        if (chromosome, strand, coordinate) in self.ajds_ids_by_positions:
            self.ajds_ids_by_positions[(chromosome, strand, coordinate)].discard(aid)

    def get_merged_adjacencies(self, merged_template="{cnt}_merged"):
        result = []
        processed_ids = set()
        cnt = 0
        for lp, rp in self.merged_position_pairs:
            lp_chr, lp_strand, lp_coord = lp
            rp_chr, rp_strand, rp_coord = rp
            lp_ids = self.get_adjs_ids_by_position(chromosome=lp_chr, strand=lp_strand, coordinate=lp_coord)
            rp_ids = self.get_adjs_ids_by_position(chromosome=rp_chr, strand=rp_strand, coordinate=rp_coord)
            common_ids = lp_ids & rp_ids
            assert len(common_ids) > 0
            assert len(common_ids & processed_ids) == 0
            processed_ids.update(common_ids)
            pos1 = Position(chromosome=lp_chr, coordinate=lp_coord, strand=lp_strand)
            pos2 = Position(chromosome=rp_chr, coordinate=rp_coord, strand=rp_strand)
            adj = Adjacency(position1=pos1, position2=pos2, extra={
                EXTERNAL_NA_ID: merged_template.format(cnt=cnt),
                self.origin_ids_field: ",".join(common_ids),
            })
            result.append(adj)
            cnt += 1
        return result

    @staticmethod
    def adjacencies_mergeable(adjacency1, adjacency2, max_distance=500, separate_distances=True):
        distance11 = Position.non_hap_distance_strand_specific(pos1=adjacency1.position1, pos2=adjacency2.position1)
        distance12 = Position.non_hap_distance_strand_specific(pos1=adjacency1.position2, pos2=adjacency2.position2)
        distance21 = Position.non_hap_distance_strand_specific(pos1=adjacency1.position2, pos2=adjacency2.position1)
        distance22 = Position.non_hap_distance_strand_specific(pos1=adjacency1.position1, pos2=adjacency2.position2)
        if not separate_distances:
            return min(distance11 + distance12, distance21 + distance22) <= max_distance
        case1 = distance11 <= max_distance and distance12 <= max_distance
        case2 = distance21 <= max_distance and distance22 <= max_distance
        return case1 or case2


def filter_adjacencies_by_chromosomal_regions(adjacencies, include=None, exclude=None, include_both=True, exclude_both=False):
    if include is None:
        include = []
    if exclude is None:
        exclude = []
    include_by_chr = defaultdict(list)
    for segment in include:
        include_by_chr[segment.chromosome.lower()].append(segment)
    exclude_by_chr = defaultdict(list)
    for segment in exclude:
        exclude_by_chr[segment.chromosome.lower()].append(segment)
    for chr in list(include_by_chr.keys()):
        include_by_chr[chr] = sorted(include_by_chr[chr], key=lambda s: (s.start_coordinate, s.end_coordinate))
    for chr in list(exclude_by_chr.keys()):
        exclude_by_chr[chr] = sorted(exclude_by_chr[chr], key=lambda s: (s.start_coordinate, s.end_coordinate))
    include_by_chr = dict(include_by_chr)
    exclude_by_chr = dict(exclude_by_chr)
    for adj in adjacencies:
        retain = True
        chr1, chr2 = adj.position1.chromosome, adj.position2.chromosome
        chr1, chr2 = chr1.lower(), chr2.lower()
        include_segments_chr1 = include_by_chr.get(chr1, [])
        include_segments_chr2 = include_by_chr.get(chr2, [])
        if len(include_segments_chr1) != 0 or len(include_segments_chr2) != 0:
            chr1in = coordinate_in_any_segment(coordinate=adj.position1.coordinate, segments=include_segments_chr1)
            chr2in = coordinate_in_any_segment(coordinate=adj.position2.coordinate, segments=include_segments_chr2)
            if include_both:
                retain &= chr1in and chr2in
            else:
                retain &= chr1in or chr2in
        elif len(include_by_chr) > 0:
            retain = False
        if not retain:
            continue
        exclude_segments_chr1 = exclude_by_chr.get(chr1, [])
        exclude_segments_chr2 = exclude_by_chr.get(chr2, [])
        if len(exclude_segments_chr1) != 0 or len(exclude_segments_chr2) != 0:
            chr1in = coordinate_in_any_segment(coordinate=adj.position1.coordinate, segments=exclude_segments_chr1)
            chr2in = coordinate_in_any_segment(coordinate=adj.position2.coordinate, segments=exclude_segments_chr2)
            if exclude_both:
                retain &= not (chr1in and chr2in)
            else:
                retain &= not (chr1in or chr2in)
        if retain:
            yield adj


def filter_adjacencies_by_size(adjacencies, min_size=0, max_size=1000000000, size_extra_field=None, size_extra_seq_field=None,
                               allow_inter_chr=True, size_extra_field_abs=True, allow_self_loops=True):
    for adj in adjacencies:
        adj_size = None
        try:
            adj_size = int(adj.extra[size_extra_field])
            if size_extra_field_abs:
                adj_size = abs(adj_size)
        except (KeyError, ValueError):
            pass
        if adj_size is None:
            try:
                adj_size = len(adj.extra[size_extra_seq_field])
            except (KeyError, ValueError):
                pass
        if adj_size is None:
            adj_size = adj.distance_non_hap
        if adj_size == 0 and allow_self_loops:
            yield adj
        elif adj_size == -1 and allow_inter_chr:
            yield adj
        elif min_size <= adj_size <= max_size:
            yield adj


def coordinate_in_any_segment(coordinate, segments):
    for segment in segments:
        if segment.start_coordinate <= coordinate <= segment.end_coordinate:
            return True
    return False


def get_shared_nas_parser():
    shared_parser = argparse.ArgumentParser(add_help=False)
    shared_parser.add_argument("--no-append-id", action="store_false", dest="append_id_suffix")
    shared_parser.add_argument("--o-extra-fields", default="all")
    shared_parser.add_argument("--chrs-include", action="append", nargs=1)
    shared_parser.add_argument("--chrs-include-file", type=argparse.FileType("rt"))
    shared_parser.add_argument("--chrs-include-no-both", action="store_false", dest="include_both")
    shared_parser.add_argument("--chrs-exclude", action="append", nargs=1)
    shared_parser.add_argument("--chrs-exclude-file", type=argparse.FileType("rt"))
    shared_parser.add_argument("--chrs-exclude-both", action="store_true", dest="exclude_both")
    return shared_parser


def get_chromosome_strip_parser():
    chr_strip_parser = argparse.ArgumentParser(add_help=False)
    chr_strip_parser.add_argument("--no-strip-chr", action="store_false", dest="strip_chr")
    return chr_strip_parser


def get_chromosome_regions_dict(string_entries):
    result = defaultdict(list)
    for entry in string_entries:
        if len(entry) == 0:
            continue
        data = entry.split(":")
        assert len(data) in [1, 2]
        chromosome = data[0]
        if len(data) > 1:
            start, stop = map(int, data[1].split("-"))
            result[chromosome].append((start, stop))
        else:
            result[chromosome].append((-1, 3000000000))
    for chromosome in result.keys():
        result[chromosome] = sorted(result[chromosome], key=lambda e: (e[0], e[1]))
    return result


def iter_over_string_entries_from_source(source):
    for line in source:
        line = line.strip()
        if len(line) == 0 or line.startswith("#"):
            continue
        yield line


def get_extra_field_regexes(string_entries):
    result = defaultdict(list)
    for entry in string_entries:
        data = entry.split("=")
        data = [data[0], "=".join(data[1:])]
        field_name = data[0]
        regex = re.compile(data[1])
        result[field_name].append(regex)
    return result


def position_in_region(position, region):
    return region[0] <= position.coordinate <= region[1]


def filter_adjacencies_by_extra(adjacencies,
                                keep_extra_field=None, keep_extra_field_missing_strategy=None,
                                remove_extra_field=None, remove_extra_field_missing_strategy=None):
    for adj in adjacencies:
        if keep_extra_field is not None and len(keep_extra_field) > 0:
            keep = False
            for field, regex_list in keep_extra_field.items():
                for regex in regex_list:
                    if field in adj.extra:
                        data = adj.extra[field]
                        if regex.search(data) is not None:
                            keep = True
                    elif keep_extra_field_missing_strategy == REMOVE:
                        keep = keep or False
            if not keep:
                continue

        if remove_extra_field is not None and len(remove_extra_field) > 0:
            keep = True
            for field, regex_list in remove_extra_field.items():
                for regex in regex_list:
                    if field in adj.extra:
                        data = adj.extra[field]
                        if regex.search(data) is not None:
                            keep = False
                            break
                    elif remove_extra_field_missing_strategy == REMOVE:
                        keep = False
                        break
            if not keep:
                continue
        yield adj


def processed_gundem2015_adjacencies(adjacencies, sample_names, min_per_sample_cnt, remove_per_sample_cnt_data=True):
    result = []
    for adj in adjacencies:
        for sample_name in sample_names:
            if GUNDEM_PER_SAMPLE_SUPPORT in adj.extra and \
                    sample_name in adj.extra[GUNDEM_PER_SAMPLE_SUPPORT] and \
                    adj.extra[GUNDEM_PER_SAMPLE_SUPPORT][sample_name] >= min_per_sample_cnt:
                result.append(adj)
                break
        if remove_per_sample_cnt_data and GUNDEM_PER_SAMPLE_SUPPORT in adj.extra:
            del adj.extra[GUNDEM_PER_SAMPLE_SUPPORT]
    return result


def refined_adjacencies_reciprocal(novel_adjacencies, max_distance, inplace=False):
    if not inplace:
        novel_adjacencies = deepcopy(novel_adjacencies)
    positions_by_chr = defaultdict(set)
    adjacencies_by_positions = defaultdict(list)
    for na in novel_adjacencies:
        p1_chr = na.position1.chromosome
        p2_chr = na.position2.chromosome
        positions_by_chr[p1_chr].add(copy(na.position1))
        positions_by_chr[p2_chr].add(copy(na.position2))
        adjacencies_by_positions[na.position1].append(na)
        adjacencies_by_positions[na.position2].append(na)
    new_positions_by_chr = {}
    for chr_name in sorted(positions_by_chr.keys()):
        new_positions_by_chr[chr_name] = sorted(positions_by_chr[chr_name], key=lambda p: p.coordinate)
    positions_by_chr = new_positions_by_chr
    for chr_name in sorted(positions_by_chr.keys()):
        positions = positions_by_chr[chr_name]
        if len(positions) < 2:
            continue
        suitable_for_merging = True
        for lp, rp in zip(positions[:-1], positions[1:]):
            if not suitable_for_merging:
                suitable_for_merging = True
                continue
            if positions_are_reciprocal(p1=lp, p2=rp, max_distance=max_distance):
                lp_adjacencies = adjacencies_by_positions[lp]
                rp_adjacencies = adjacencies_by_positions[rp]
                lp_adjacencies_ids = {a.stable_id_non_phased for a in lp_adjacencies}
                rp_adjacencies_ids = {a.stable_id_non_phased for a in rp_adjacencies}
                if len(lp_adjacencies_ids & rp_adjacencies_ids) > 0:
                    continue
                new_p1, new_p2 = merged_reciprocal_positions(p1=lp, p2=rp)
                if new_p1.strand == rp.strand:
                    new_p1, new_p2 = new_p2, new_p1
                for adj in lp_adjacencies:
                    update_position_in_adjacency(adjacency=adj, old_position=lp, new_position=new_p1)
                for adj in rp_adjacencies:
                    update_position_in_adjacency(adjacency=adj, old_position=rp, new_position=new_p2)
                suitable_for_merging = False
    return novel_adjacencies


def positions_are_reciprocal(p1, p2, max_distance):
    if p1.chromosome != p2.chromosome:
        return False
    return abs(p1.coordinate - p2.coordinate) <= max_distance and p1.strand != p2.strand


def merged_reciprocal_positions(p1, p2):
    p1, p2 = deepcopy(p1), deepcopy(p2)
    fsp, rsp = (p1, p2) if p1.strand == Strand.FORWARD else (p2, p1)
    left = min([p1.coordinate, p2.coordinate])
    distance = abs(p1.coordinate - p2.coordinate)
    mid_coordinate = left + int(distance / 2)
    fsp.coordinate = mid_coordinate
    rsp.coordinate = mid_coordinate + 1
    return fsp, rsp


def update_position_in_adjacency(adjacency, old_position, new_position):
    if adjacency.position1 == old_position:
        adjacency.position1 = deepcopy(new_position)
    if adjacency.position2 == old_position:
        adjacency.position2 = deepcopy(new_position)



import argparse
import sys
import os

current_file_level = 3
current_dir = os.path.dirname(os.path.realpath(__file__))
for _ in range(current_file_level):
    current_dir = os.path.dirname(current_dir)
sys.path.append(current_dir)

from rck.core.io import get_logging_cli_parser, get_standard_logger_from_args, get_full_path, read_scnt_from_file, write_scnt_to_file, \
    write_scnt_to_destination, read_scnt_from_source
from rck.core.structures import aligned_scnts, refined_scnt, cn_distance_inter_scnt


def main():
    parser = argparse.ArgumentParser(prog="RCK-UTILS-SCNT-process")
    cli_logging_parser = get_logging_cli_parser()
    subparsers = parser.add_subparsers(title="command", dest="command")
    subparsers.required = True
    ###
    refine_parser = subparsers.add_parser("refine", parents=[cli_logging_parser])
    refine_parser.add_argument('scnt', type=argparse.FileType("rt"), default=sys.stdin)
    refine_parser.add_argument("--separator", default="\t")
    refine_parser.add_argument("--no-allow-missing-clones", action="store_false", dest="allow_missing_clones")
    refine_parser.add_argument("--clone-ids", default=None)
    refine_parser.add_argument("--no-merge-fragments", action="store_false", dest="merge_fragments")
    refine_parser.add_argument("--max-merge-gap", type=int, default=1000000)
    refine_parser.add_argument("--no-fill-gaps", action="store_false", dest="fill_gaps")
    refine_parser.add_argument("--max-fill-gap", type=int, default=1000000)
    refine_parser.add_argument('--output', type=argparse.FileType("wt"), default=sys.stdout)
    ###
    align_parser = subparsers.add_parser("align", parents=[cli_logging_parser])
    align_parser.add_argument("scnt", nargs="+")
    align_parser.add_argument("--separator", default="\t")
    align_parser.add_argument("--output-suffix", default="aligned")
    align_parser.add_argument("--no-allow-unit-segments", action="store_false", dest="allow_unit_segments")
    align_parser.add_argument("--output-dir", default="")
    ###
    distance_parser = subparsers.add_parser("distance", parents=[cli_logging_parser])
    distance_parser.add_argument("--scnt1", type=argparse.FileType("rt"), required=True)
    distance_parser.add_argument("--scnt1-separator", default="\t")
    distance_parser.add_argument("--scnt1-extra-separator", default=";")
    distance_parser.add_argument("--scnt2", type=argparse.FileType("rt"), required=True)
    distance_parser.add_argument("--scnt2-separator", default="\t")
    distance_parser.add_argument("--scnt2-extra-separator", default=";")
    distance_parser.add_argument("--clone-ids", default=None)
    distance_parser.add_argument("--output", "-o", type=argparse.FileType("wt"), default=sys.stdout)
    ###
    args = parser.parse_args()
    logger = get_standard_logger_from_args(args=args, program_name="RCK-UTILS-SCNT-process")

    if args.command == "refine":
        clone_ids = args.clone_ids.split(",") if args.clone_ids is not None else None
        logger.debug("Clone ids identified as {clone_ids}. If None -- all clone ids will be processed.".format(clone_ids=",".join(clone_ids)))
        logger.info("Reading Segment Copy Number Tensor form {file}".format(file=args.scnt))
        segments, scnt = read_scnt_from_source(source=args.scnt, clone_ids=clone_ids, separator=args.separator)
        logger.info("Refining Segment Copy Number Tensor from {file}".format(file=args.scnt))
        segments, scnt, _ = refined_scnt(segments=segments, scnt=scnt,
                                         merge_fragments=args.merge_fragments, max_merge_gap=args.max_merge_gap,
                                         fill_gaps=args.fill_gaps, max_fill_gap=args.max_fill_gap)
        logger.info("Writing refined Segment Copy Number Tensor to {file}".format(file=args.output))
        write_scnt_to_destination(destination=args.output, scnt=scnt, segments=segments, clone_ids=clone_ids, separator=args.separator)
    elif args.command == "align":
        scnt_files = {}
        for path in args.scnt:
            full_path = get_full_path(path=path)
            name = os.path.splitext(os.path.basename(full_path))[0]
            if name.endswith(".scnt"):
                name = name[:-5]
            if name.endswith("."):
                name = name[:-1]
            scnt_files[name] = full_path
        logger.debug("Input Segment Copy Number Tensors (SCNT) identified as {input_scnts}".format(input_scnts=" , ".join(scnt_files.values())))
        scnts_by_name = {}
        segments_by_name = {}
        clone_ids_by_scnt = {}
        logger.info("Reading input SCNTs")
        for name, path in scnt_files.items():
            logger.debug("Reading SCNT from {file}".format(file=scnt_files[name]))
            segments, scnt = read_scnt_from_file(file_name=scnt_files[name], separator=args.separator)
            clone_ids_by_scnt[name] = sorted(scnt.keys())
            scnts_by_name[name] = scnt
            segments_by_name[name] = segments
        if len(scnts_by_name.values()) == 1:
            logger.warning("Only one input SCNT identified. Doing nothing with it, outputting as is.")
            aligned_segments_by_name, aligned_scnts_by_name = segments_by_name, scnts_by_name
        else:
            logger.info("Aligning input SCNTs.")
            aligned_segments_by_name, aligned_scnts_by_name = aligned_scnts(segments_by_sample_names=segments_by_name, scnts_by_sample_names=scnts_by_name)
        result_base_names = {}
        cnt = 0
        for name in sorted(scnt_files.keys()):
            new_name = name
            if name in result_base_names:
                new_name = name + str(cnt)
                cnt += 1
            new_name = new_name + "." + args.output_suffix
            result_base_names[name] = new_name
        output_dir = args.output_dir if args.output_dir != "" else os.getcwd()
        output_dir = get_full_path(path=output_dir)
        logger.info("Writing aligned SCNTs")
        for name, new_name in result_base_names.items():
            scnt = aligned_scnts_by_name[name]
            segments = aligned_segments_by_name[name]
            scnt_path = os.path.join(output_dir, new_name + "rck.scnt.tsv")
            logger.debug("Writing aligned SCNT {scnt_name} to {file}".format(scnt_name=name, file=scnt_path))
            write_scnt_to_file(file_name=scnt_path, segments=segments, scnt=scnt, separator=args.separator)
    elif args.command == "distance":
        clone_ids = args.clone_ids
        if args.clone_ids is not None:
            clone_ids = args.clone_ids.split(",")
        segments1, scnt1 = read_scnt_from_source(source=args.scnt1, clone_ids=clone_ids, separator=args.scnt1_separator,
                                                 extra_separator=args.scnt1_extra_separator, remove_cn_data_from_segs=True)
        segments2, scnt2 = read_scnt_from_source(source=args.scnt2, clone_ids=clone_ids, separator=args.scnt2_separator,
                                                 extra_separator=args.scnt2_extra_separator, remove_cn_data_from_segs=True)
        segments_by_sample_names = {"1": segments1, "2": segments2}
        scnts_by_sample_names = {"1": scnt1, "2": scnt2}
        segments_by_sample_names, scnts_by_sample_names = aligned_scnts(segments_by_sample_names=segments_by_sample_names,
                                                                        scnts_by_sample_names=scnts_by_sample_names)
        segments = segments_by_sample_names["1"]
        scnt1, scnt2 = scnts_by_sample_names["1"], scnts_by_sample_names["2"]
        distance = cn_distance_inter_scnt(tensor1=scnt1, tensor2=scnt2, segments=segments, check_clone_ids_match=True)
        print("distance = ", distance)

    logger.info("Success!")


if __name__ == "__main__":
    pass



import argparse

import os

import sys


current_file_level = 3
current_dir = os.path.dirname(os.path.realpath(__file__))
for _ in range(current_file_level):
    current_dir = os.path.dirname(current_dir)
sys.path.append(current_dir)

import rck
from rck.core.io import read_scnt_from_source, extract_scnb_from_segments, write_scnb_to_destination
from rck.core.structures import SCNBoundariesStrategies, LengthSpreadRelationships, SegmentCopyNumberBoundaries


def main():
    parser = argparse.ArgumentParser("Creating boundaries for a RCK formatted segment copy number tensor")
    parser.add_argument('--version', action='version', version=rck.version)
    parser.add_argument("scnt", type=argparse.FileType("rt"), default=sys.stdin)
    parser.add_argument("--bnd-strategy", choices=[strategy.value for strategy in SCNBoundariesStrategies], type=SCNBoundariesStrategies.from_string,
                        default=SCNBoundariesStrategies.UNIFORM_MIN_MAX.value)
    parser.add_argument("--uniform-spread-size", type=int, default=1)
    parser.add_argument("--length-spread-relation", choices=[rel.value for rel in LengthSpreadRelationships], type=LengthSpreadRelationships.from_string,
                        default=LengthSpreadRelationships.DUMMY.value)
    parser.add_argument("--uniform-min", type=int, default=0)
    parser.add_argument("--uniform-max", type=int, default=10)
    parser.add_argument("--missing-only", action="store_true", dest="missing_only")
    parser.add_argument("--min-allow-zero-for-positive", type=int, default=-1)
    parser.add_argument("--max-allow-zero-for-positive", type=int, default=1000000000)
    parser.add_argument("--min-allow-positive-for-zero", type=int, default=-1)
    parser.add_argument("--max-allow-positive-for-zero", type=int, default=1000000000)
    parser.add_argument("--clone-ids", default="")
    parser.add_argument("--is-male", action="store_false", dest="is_female")
    parser.add_argument("--no-allow-unit-segments", dest="allow_unit_segments", action="store_false")
    parser.add_argument("--output", "-o", type=argparse.FileType("wt"), default=sys.stdout)
    parser.add_argument("--o-with-scnt", action="store_true", dest="output_scnt")
    parser.add_argument("--separator", default="\t")
    args = parser.parse_args()
    clone_ids = args.clone_ids.split(",") if len(args.clone_ids) > 0 else None
    segments, scnt = read_scnt_from_source(source=args.scnt, clone_ids=clone_ids)
    if clone_ids is None:
        clone_ids = sorted(scnt.keys())
    try:
        scnb = extract_scnb_from_segments(segments=segments, clone_ids=clone_ids)
    except ValueError:
        scnb = {clone_id: SegmentCopyNumberBoundaries() for clone_id in clone_ids}
    for clone_id in clone_ids:
        scnb[clone_id].fill(segments=segments, scnp=scnt[clone_id], missing_only=args.missing_only, strategy=args.bnd_strategy,
                            min_allow_zero_for_positive=args.min_allow_zero_for_positive,
                            max_allow_zero_for_positive=args.max_allow_zero_for_positive,
                            min_allow_positive_for_zero=args.min_allow_positive_for_zero,
                            max_allow_positive_for_zero=args.max_allow_positive_for_zero,
                            uniform_spread_size=args.uniform_spread_size,
                            length_spread_relation=args.length_spread_relation,
                            uniform_min=args.uniform_min,
                            uniform_max=args.uniform_max,
                            is_female=args.is_female)
    write_scnb_to_destination(destination=args.output, segments=segments, scnb=scnb, clone_ids=clone_ids)


if __name__ == "__main__":
    main()



import csv
from collections import defaultdict
from copy import deepcopy

import vcf

from rck.core.io import EXTERNAL_NA_ID, SVTYPE, write_adjacencies_to_destination, COPY_NUMBER
from rck.core.structures import Strand, Position, Adjacency, AdjacencyType, Phasing, Segment

STRIP_CHR = "strip_CHR"
APPEND_ID = "append_id"
EXTRA_FIELDS = "extra_fields"
ID_SUFFIX = "id_suffix"
SVTYPE2 = "svtype2"
EXTERNAL_NA_ID2 = "external_na_id2"
SVLEN = "svlen"
SUPPORT_READ_NAMES = "support_read_names"


def get_most_supported_strands(strands_list):
    result_by_support = {}
    for entry in strands_list:
        strands, support = entry.split(":")
        support = int(support)
        result_by_support[support] = strands
    max_support = max(result_by_support.keys())
    return result_by_support[max_support]


def strip_chr(chr_string):
    if chr_string.startswith("chr"):
        return chr_string[3:]
    return chr_string


def get_strand_string_from_alt_breakend_string(alt_breakend_string):
    if alt_breakend_string.startswith("]") or alt_breakend_string.endswith("]"):
        return "+"
    return "-"


def get_strand_string_from_alt_breakend(alt_breakend):
    return get_strand_string_from_alt_breakend_string(alt_breakend_string=str(alt_breakend))


def update_dict_with_vcf_info(target, vcf_record):
    for key, value in vcf_record:
        target[str(key).lower()] = value
    return target


def get_inv_chrs_from_vcf_record(vcf_record):
    chr11 = str(vcf_record.CHROM)
    chr12 = chr11
    chr21 = chr11
    chr22 = chr11
    return chr11, chr12, chr21, chr22


def get_inv_na1_chrs_from_vcf_record(vcf_record):
    return str(vcf_record.CHROM), str(vcf_record.CHROM)


def get_inv_na2_chrs_from_vcf_record(vcf_record):
    return get_inv_na1_chrs_from_vcf_record(vcf_record=vcf_record)


def get_inv_na1_coords_from_vcf_record(vcf_record):
    return int(vcf_record.POS) - 1, int(vcf_record.INFO["END"])


def get_inv_na2_coords_from_vcf_record(vcf_record):
    a, b = get_inv_na1_coords_from_vcf_record(vcf_record=vcf_record)
    return a + 1, b + 1


def get_inv_na1_strands_from_vcf_record(vcf_record):
    return Strand.FORWARD, Strand.FORWARD


def get_inv_na2_strands_from_vcf_record(vcf_record):
    return Strand.REVERSE, Strand.REVERSE


def get_del_na_chrs_from_vcf_record(vcf_record):
    return get_inv_na1_chrs_from_vcf_record(vcf_record=vcf_record)


def get_del_na_coords_from_vcf_record(vcf_record):
    return int(vcf_record.POS) - 1, int(vcf_record.INFO["END"]) + 1


def get_del_na_strands_from_vcf_record(vcf_record):
    return Strand.FORWARD, Strand.REVERSE


def get_dup_na_chrs_from_vcf_record(vcf_record):
    return get_inv_na1_chrs_from_vcf_record(vcf_record=vcf_record)


def get_dup_na_coords_from_vcf_record(vcf_record):
    return int(vcf_record.POS), int(vcf_record.INFO["END"])


def get_dup_na_strands_from_vcf_record(vcf_record):
    return Strand.REVERSE, Strand.FORWARD


def get_strand_from_alt_breakend(vcf_recrod):
    alt_string = str(vcf_recrod.ALT[0])
    if alt_string.startswith("]") or alt_string.endswith("]"):
        return Strand.FORWARD
    return Strand.REVERSE


def get_vcf_records_from_file(vcf_file_name):
    with open(vcf_file_name, "rt") as source:
        return get_vcf_records_from_source(source=source)


def get_vcf_records_from_source(source):
    result = []
    vcf_reader = vcf.Reader(source)
    if "RNAMES" in vcf_reader.infos:
        vcf_reader.infos["RNAMES"] = vcf_reader.infos["RNAMES"]._replace(num='.')
    for record in vcf_reader:
        result.append(record)
    return result


def get_vcf_records_by_ids(vcf_records):
    return {r.ID: r for r in vcf_records}


def get_nas_from_lumpy_vcf_file(lumpy_vcf_file):
    vcf_records = get_vcf_records_from_file(vcf_file_name=lumpy_vcf_file)
    return get_nas_from_lumpy_vcf_records(lumpy_vcf_records=vcf_records)


def update_nas_ids(nas_by_ids_defaultdict, setup):
    result = {}
    for sv_id, nas in nas_by_ids_defaultdict.items():
        if len(nas) > 1:
            for cnt, na in enumerate(nas, start=1):
                na.extra[EXTERNAL_NA_ID] += "_{cnt}".format(cnt=cnt)
        if setup.get(APPEND_ID, True):
            for na in nas:
                if not na.extra[EXTERNAL_NA_ID].endswith(setup.get(ID_SUFFIX, "")):
                    na.extra[EXTERNAL_NA_ID] += "_{suffix}".format(suffix=setup.get(ID_SUFFIX, ""))
        for na in nas:
            na.extra[EXTERNAL_NA_ID] = na.extra[EXTERNAL_NA_ID].replace(":", "_")
            result[na.extra[EXTERNAL_NA_ID]] = na
    return result


DUPLICATED_ENTRIES_EXTRA = {
    "CHR2",
    "END",
    "STRANDS"
}


def clear_duplicated_entries_extra(extra):
    to_remove = []
    for key, value in extra.items():
        if key.upper() in DUPLICATED_ENTRIES_EXTRA:
            to_remove.append(key)
    for key in to_remove:
        extra.pop(key)
    return extra


def get_nas_from_lumpy_vcf_records(lumpy_vcf_records, setup=None):
    if setup is None:
        setup = {}
    records_by_ids = get_vcf_records_by_ids(vcf_records=lumpy_vcf_records)
    nas_by_ids = defaultdict(list)
    processed_vcf_ids = set()
    for record in records_by_ids.values():
        base_sv_id = str(record.ID).split("_")[0]
        if base_sv_id in processed_vcf_ids:
            continue
        sv_type = record.INFO["SVTYPE"]
        extra = {}
        if sv_type in ["DUP", "DEL", "BND"]:
            if sv_type in ["DUP", "DEL"]:
                if "_" in str(record.ID):
                    raise ValueError("Non standard id {sv_id} for a same-chromosome SV of type {sv_type}".format(sv_id=record.ID, sv_type=sv_type))
                extra.update(record.INFO)
                chr1, chr2 = get_dup_na_chrs_from_vcf_record(vcf_record=record)
                if sv_type == "DUP":
                    strand1, strand2 = get_dup_na_strands_from_vcf_record(vcf_record=record)
                    coord1, coord2 = get_dup_na_coords_from_vcf_record(vcf_record=record)
                else:
                    strand1, strand2 = get_del_na_strands_from_vcf_record(vcf_record=record)
                    coord1, coord2 = get_del_na_coords_from_vcf_record(vcf_record=record)
            else:
                if "_" not in str(record.ID):
                    raise ValueError("Non standard id {sv_id} for a BND SV".format(sv_id=record.ID))
                extra.update(record.INFO)
                chr1 = str(record.CHROM)
                coord1 = int(record.POS)
                mate_vcf_entry_id_list = record.INFO["MATEID"]
                mate_vcf_entry_id = mate_vcf_entry_id_list[0]
                mate_record = records_by_ids[mate_vcf_entry_id]
                extra.update(mate_record.INFO)
                strand1 = get_strand_from_alt_breakend(vcf_recrod=mate_record)
                strand2 = get_strand_from_alt_breakend(vcf_recrod=record)
                chr2 = str(mate_record.CHROM)
                coord2 = int(mate_record.POS)
            if setup.get(STRIP_CHR, True):
                chr1 = strip_chr(chr_string=chr1)
                chr2 = strip_chr(chr_string=chr2)
            pos1 = Position(chromosome=chr1, coordinate=coord1, strand=strand1)
            pos2 = Position(chromosome=chr2, coordinate=coord2, strand=strand2)
            extra[EXTERNAL_NA_ID] = base_sv_id
            extra = clear_duplicated_entries_extra(extra=extra)
            na = Adjacency(position1=pos1, position2=pos2, extra=extra)
            nas_by_ids[base_sv_id].append(na)
            processed_vcf_ids.add(base_sv_id)
        elif sv_type == "INV":
            if "_" in str(record.ID):
                raise Exception("Non standard id {sv_id} for a same-chromosome SV of type {sv_type}".format(sv_id=record.ID, sv_type=sv_type))
            extra1 = deepcopy(record.INFO)
            extra2 = deepcopy(record.INFO)
            chr11, chr12 = get_inv_na1_chrs_from_vcf_record(vcf_record=record)
            chr21, chr22 = get_inv_na2_chrs_from_vcf_record(vcf_record=record)
            coord11, coord12 = get_inv_na1_coords_from_vcf_record(vcf_record=record)
            coord21, coord22 = get_inv_na2_coords_from_vcf_record(vcf_record=record)
            strand11, strand12 = get_inv_na1_strands_from_vcf_record(vcf_record=record)
            strand21, strand22 = get_inv_na2_strands_from_vcf_record(vcf_record=record)
            if setup.get(STRIP_CHR, True):
                chr11 = strip_chr(chr_string=chr11)
                chr12 = strip_chr(chr_string=chr12)
                chr21 = strip_chr(chr_string=chr21)
                chr22 = strip_chr(chr_string=chr22)
            pos11 = Position(chromosome=chr11, coordinate=coord11, strand=strand11)
            pos12 = Position(chromosome=chr12, coordinate=coord12, strand=strand12)
            pos21 = Position(chromosome=chr21, coordinate=coord21, strand=strand21)
            pos22 = Position(chromosome=chr22, coordinate=coord22, strand=strand22)
            extra1[EXTERNAL_NA_ID] = base_sv_id
            extra2[EXTERNAL_NA_ID] = base_sv_id
            extra1 = clear_duplicated_entries_extra(extra=extra1)
            extra2 = clear_duplicated_entries_extra(extra=extra2)
            na1 = Adjacency(position1=pos11, position2=pos12, extra=extra1)
            na2 = Adjacency(position1=pos21, position2=pos22, extra=extra2)
            nas_by_ids[base_sv_id].append(na1)
            nas_by_ids[base_sv_id].append(na2)
            processed_vcf_ids.add(base_sv_id)
        else:
            raise Exception("Unknown SVTYPE {sv_type}".format(sv_type=sv_type))
    nas_by_ids = update_nas_ids(nas_by_ids_defaultdict=nas_by_ids, setup=setup)
    return list(nas_by_ids.values())


def get_nas_from_longranger_vcf_records(longranger_vcf_records, setup=None):
    if setup is None:
        setup = {}
    records_by_ids = get_vcf_records_by_ids(vcf_records=longranger_vcf_records)
    nas_by_ids = defaultdict(list)
    processed_vcf_ids = set()
    for record in records_by_ids.values():
        sv_id_entries = str(record.ID).split("_")
        sv_id = str(record.ID)
        if sv_id in processed_vcf_ids:
            continue
        extra = deepcopy(record.INFO)
        sv_type = record.INFO["SVTYPE"]
        if len(sv_id_entries) <= 2:
            mate_present = False
        else:
            mate_present = True
        assert mate_present == ("MATEID" in record.INFO)
        if mate_present:
            base_sv_id = "_".join(sv_id_entries[:-1])
            for mate_vcf_id in record.INFO["MATEID"]:
                mate_record = records_by_ids[mate_vcf_id]
                extra.update(deepcopy(mate_record.INFO))
                mate_breakend = record.ALT[0]
                record_breakend = mate_record.ALT[0]
                chr1 = str(record.CHROM)
                chr2 = str(mate_record.CHROM)
                strand1 = Strand.from_pm_string(get_strand_string_from_alt_breakend(record_breakend))
                strand2 = Strand.from_pm_string(get_strand_string_from_alt_breakend(mate_breakend))
                coord1 = int(record.POS)
                coord2 = int(mate_record.POS)
                if setup.get(STRIP_CHR, True):
                    chr1 = strip_chr(chr_string=chr1)
                    chr2 = strip_chr(chr_string=chr2)
                pos1 = Position(chromosome=chr1, coordinate=coord1, strand=strand1)
                pos2 = Position(chromosome=chr2, coordinate=coord2, strand=strand2)
                extra[EXTERNAL_NA_ID] = base_sv_id
                na = Adjacency(position1=pos1, position2=pos2, extra=extra)
                nas_by_ids[base_sv_id].append(na)
                processed_vcf_ids.add(mate_vcf_id)
            processed_vcf_ids.add(sv_id)
        else:
            if sv_type == "UNK":
                continue
            base_sv_id = "_".join(sv_id_entries)
            if sv_type == "INV":
                chr11, chr12 = get_inv_na1_chrs_from_vcf_record(vcf_record=record)
                chr21, chr22 = get_inv_na2_chrs_from_vcf_record(vcf_record=record)
                coord11, coord12 = get_inv_na1_coords_from_vcf_record(vcf_record=record)
                coord21, coord22 = get_inv_na2_coords_from_vcf_record(vcf_record=record)
                strand11, strand12 = get_inv_na1_strands_from_vcf_record(vcf_record=record)
                strand21, strand22 = get_inv_na2_strands_from_vcf_record(vcf_record=record)
                if setup.get(STRIP_CHR, True):
                    chr11 = strip_chr(chr_string=chr11)
                    chr12 = strip_chr(chr_string=chr12)
                    chr21 = strip_chr(chr_string=chr21)
                    chr22 = strip_chr(chr_string=chr22)
                pos11 = Position(chromosome=chr11, coordinate=coord11, strand=strand11)
                pos12 = Position(chromosome=chr12, coordinate=coord12, strand=strand12)
                pos21 = Position(chromosome=chr21, coordinate=coord21, strand=strand21)
                pos22 = Position(chromosome=chr22, coordinate=coord22, strand=strand22)
                extra[EXTERNAL_NA_ID] = base_sv_id
                extra1 = deepcopy(extra)
                extra2 = deepcopy(extra)
                na1 = Adjacency(position1=pos11, position2=pos12, extra=extra1)
                na2 = Adjacency(position1=pos21, position2=pos22, extra=extra2)
                nas_by_ids[base_sv_id].append(na1)
                nas_by_ids[base_sv_id].append(na2)
                processed_vcf_ids.add(base_sv_id)
            elif sv_type in ["DEL", "DUP"]:
                chr1 = str(record.CHROM)
                if setup.get(STRIP_CHR, True):
                    chr1 = strip_chr(chr_string=chr1)
                chr2 = chr1
                coord1 = int(record.POS)
                coord2 = int(record.INFO["END"])
                if sv_type == "DUP":
                    strand1 = Strand.REVERSE
                    strand2 = Strand.FORWARD
                else:
                    strand1 = Strand.FORWARD
                    strand2 = Strand.REVERSE
                extra[EXTERNAL_NA_ID] = base_sv_id
                pos1 = Position(chromosome=chr1, coordinate=coord1, strand=strand1)
                pos2 = Position(chromosome=chr2, coordinate=coord2, strand=strand2)
                na = Adjacency(position1=pos1, position2=pos2, extra=extra)
                nas_by_ids[base_sv_id].append(na)
                processed_vcf_ids.add(base_sv_id)
            else:
                raise Exception("Unknown SV type ({svtype}) for longranger".format(svtype=sv_type))
    nas_by_ids = update_nas_ids(nas_by_ids_defaultdict=nas_by_ids, setup=setup)
    return list(nas_by_ids.values())


def get_nas_from_manta_vcf_records(manta_vcf_records, setup=None):
    nas_by_ids = defaultdict(list)
    if setup is None:
        setup = {}
    records_by_ids = get_vcf_records_by_ids(vcf_records=manta_vcf_records)
    processed_ids = set()
    for record in records_by_ids.values():
        extra = deepcopy(record.INFO)
        svtype = record.INFO["SVTYPE"]
        record_id = str(record.ID)
        if record_id in processed_ids:
            continue
        if svtype == "BND":
            mate_vcf_id = record.INFO["MATEID"][0]
            if mate_vcf_id not in records_by_ids:
                continue
            mate_record = records_by_ids[mate_vcf_id]
            chr1 = str(record.CHROM)
            chr2 = str(mate_record.CHROM)
            coord1 = int(record.POS)
            coord2 = int(mate_record.POS)
            strand1 = Strand.from_pm_string(string=get_strand_string_from_alt_breakend(alt_breakend=mate_record.ALT))
            strand2 = Strand.from_pm_string(string=get_strand_string_from_alt_breakend(alt_breakend=record.ALT))
            if setup.get(STRIP_CHR, True):
                chr1, chr2 = strip_chr(chr_string=chr1), strip_chr(chr_string=chr2)
            pos1 = Position(chromosome=chr1, coordinate=coord1, strand=strand1)
            pos2 = Position(chromosome=chr2, coordinate=coord2, strand=strand2)
            extra[EXTERNAL_NA_ID] = str(record.ID)
            extra[EXTERNAL_NA_ID2] = mate_vcf_id
            extra = clear_duplicated_entries_extra(extra=extra)
            na = Adjacency(position1=pos1, position2=pos2, extra=extra)
            nas_by_ids[na.extra[EXTERNAL_NA_ID]].append(na)
            processed_ids.add(record_id)
            processed_ids.add(mate_vcf_id)
        elif svtype in ["INV", "DUP", "DEL", "INS"]:
            chr1 = str(record.CHROM)
            if setup.get(STRIP_CHR, True):
                chr1 = strip_chr(chr_string=chr1)
            chr2 = chr1
            coord1 = int(record.POS)
            coord2 = int(record.INFO["END"])
            if svtype == "INV":
                if "INV3" in record.INFO:
                    strand1 = Strand.REVERSE
                    strand2 = Strand.REVERSE
                elif "INV5" in record.INFO:
                    strand1 = Strand.FORWARD
                    strand2 = Strand.FORWARD
                else:
                    raise Exception("Unknown INV type for Manta (i.e., not INV3, nor INV5)")
            elif svtype == "DUP":
                strand1 = Strand.REVERSE
                strand2 = Strand.FORWARD
            elif svtype == "DEL":
                strand1 = Strand.FORWARD
                strand2 = Strand.REVERSE
            else:
                # insertion
                strand1 = Strand.FORWARD
                strand2 = Strand.REVERSE
            pos1 = Position(chromosome=chr1, coordinate=coord1, strand=strand1)
            pos2 = Position(chromosome=chr2, coordinate=coord2, strand=strand2)
            extra[EXTERNAL_NA_ID] = record_id
            if svtype == "INS":
                if "SVLEN" in record.INFO:
                    extra[SVLEN] = int(record.INFO["SVLEN"][0])
                    del extra["SVLEN"]
            na = Adjacency(position1=pos1, position2=pos2, extra=extra)
            nas_by_ids[record_id].append(na)
            processed_ids.add(record_id)
        else:
            raise Exception("Unknown SV type {svtype}".format(svtype=svtype))
    nas_by_ids = update_nas_ids(nas_by_ids_defaultdict=nas_by_ids, setup=setup)
    return nas_by_ids.values()


def get_nas_from_sniffles_vcf_records(sniffles_vcf_records, setup=None):
    if setup is None:
        setup = {}
    nas_by_ids = defaultdict(list)
    records_by_ids = get_vcf_records_by_ids(vcf_records=sniffles_vcf_records)
    processed_ids = set()
    for record in records_by_ids.values():
        extra = deepcopy(record.INFO)
        record_id = str(record.ID)
        if record_id in processed_ids:
            continue
        svtype = record.INFO.get("SVTYPE", "")
        chr1 = record.CHROM
        coord1 = int(record.POS)
        if "CHR2" in record.INFO:
            chr2 = record.INFO["CHR2"]
            coord2 = int(record.INFO["END"])
        else:
            record_breakend = record.ALT[0]
            chr2 = record_breakend.chr
            coord2 = record_breakend.pos
        try:
            strands = record.INFO["STRANDS"]
            if isinstance(strands, (list, tuple)):
                strands = strands[0]
        except KeyError:
            if svtype == "INS":
                strands = "+-"
            else:
                raise ValueError("Unknown strands {missing STRANDS entry}")
        strand1 = Strand.from_pm_string(string=strands[0])
        strand2 = Strand.from_pm_string(string=strands[1])
        if setup.get(STRIP_CHR, True):
            chr1, chr2 = strip_chr(chr_string=chr1), strip_chr(chr_string=chr2)
        extra.update({
            EXTERNAL_NA_ID: record_id,
        })
        extra = clear_duplicated_entries_extra(extra=extra)
        pos1 = Position(chromosome=chr1, coordinate=coord1, strand=strand1)
        pos2 = Position(chromosome=chr2, coordinate=coord2, strand=strand2)
        na = Adjacency(position1=pos1, position2=pos2, extra=extra)
        nas_by_ids[record_id].append(na)
        processed_ids.add(record_id)
    nas_by_ids = update_nas_ids(nas_by_ids_defaultdict=nas_by_ids, setup=setup)
    return nas_by_ids.values()


def get_nas_from_grocsv_vcf_records(grocsv_vcf_records, setup=None, samples=None, sample_all_any="any", samples_only=False):
    if setup is None:
        setup = {}
    nas_by_ids = defaultdict(list)
    records_by_ids = get_vcf_records_by_ids(vcf_records=grocsv_vcf_records)
    processed_ids = set()
    for record in records_by_ids.values():
        extra = deepcopy(record.INFO)
        record_id = str(record.ID)
        if record_id in processed_ids:
            continue
        svtype = record.INFO["SVTYPE"]
        if svtype == "BND":
            mate_vcf_record_id = record.INFO["MATEID"]
            svid = record.INFO["EVENT"]
            mate_vcf_record = records_by_ids[mate_vcf_record_id]
            processed_ids.add(record_id)
            processed_ids.add(mate_vcf_record_id)
            if samples is not None:
                present = {}
                for format_sample in record.samples:
                    present[format_sample.sample] = str(format_sample.data.GT) in ["1", "."]
                refute = False
                if samples_only:
                    refute = any([value for key, value in present.items() if key not in samples])
                if samples_all_any == "any":
                    present = any([present.get(sample_name, True) for sample_name in samples])
                else:
                    present = all([present.get(sample_name, True) for sample_name in samples])
                present &= not refute
                if not present:
                    continue
            extra.update(mate_vcf_record.INFO)
            chr1 = record.CHROM
            chr2 = mate_vcf_record.CHROM
            coord1 = int(record.POS)
            coord2 = int(mate_vcf_record.POS)
            strand1 = get_strand_from_alt_breakend(vcf_recrod=mate_vcf_record)
            strand2 = get_strand_from_alt_breakend(vcf_recrod=record)
            if setup.get(STRIP_CHR, True):
                chr1, chr2 = strip_chr(chr_string=chr1), strip_chr(chr_string=chr2)
            pos1 = Position(chromosome=chr1, coordinate=coord1, strand=strand1)
            pos2 = Position(chromosome=chr2, coordinate=coord2, strand=strand2)
            extra[EXTERNAL_NA_ID] = svid
            na = Adjacency(position1=pos1, position2=pos2, extra=extra)
            nas_by_ids[svid].append(na)
        else:
            raise Exception("Unknown SV type \"{svtype}\" for GROCSVS VCF entry".format(svtype=svtype))
    nas_by_ids = update_nas_ids(nas_by_ids_defaultdict=nas_by_ids, setup=setup)
    return nas_by_ids.values()


def get_nas_from_naibr_file(naibr_file_name, setup=None):
    if setup is None:
        setup = {}
    with open(naibr_file_name, "rt") as source:
        return get_nas_from_naibr_source(source=source, setup=setup)


def get_nas_from_naibr_source(source, setup=None):
    if setup is None:
        setup = {}
    result = defaultdict(list)
    csv_reader = csv.DictReader(source, delimiter="\t")
    for cnt, row in enumerate(csv_reader):
        chr1 = row["Chr1"]
        chr2 = row["Chr2"]
        if setup.get(STRIP_CHR, True):
            chr1 = strip_chr(chr_string=chr1)
            chr2 = strip_chr(chr_string=chr2)
        coord1 = int(row["Break1"])
        coord2 = int(row["Break2"])
        strands = row["Orientation"]
        strand1 = Strand.from_pm_string(string=strands[0])
        strand2 = Strand.from_pm_string(string=strands[1])
        pos1 = Position(chromosome=chr1, coordinate=coord1, strand=strand1)
        pos2 = Position(chromosome=chr2, coordinate=coord2, strand=strand2)
        na_id = "{cnt}".format(cnt=cnt)
        na = Adjacency(position1=pos1, position2=pos2, extra={EXTERNAL_NA_ID: na_id})
        result[na_id].append(na)
    result = update_nas_ids(nas_by_ids_defaultdict=result, setup=setup)
    return list(result.values())


def get_nas_from_delly_vcf_records(delly_vcf_records, setup=None):
    if setup is None:
        setup = {}
    result = defaultdict(list)
    records_by_ids = get_vcf_records_by_ids(vcf_records=delly_vcf_records)
    for record in records_by_ids.values():
        extra = deepcopy(record.INFO)
        svtype = record.INFO["SVTYPE"]
        svid = str(record.ID)
        extra[SVTYPE] = svtype
        extra[EXTERNAL_NA_ID] = svid
        if svtype in ["DEL", "DUP", "INV", "INS", "BND"]:
            chr1 = record.CHROM
            coord1 = int(record.POS)
            chr2 = record.INFO["CHR2"]
            coord2 = int(record.INFO["END"])
            ct_string = record.INFO["CT"]
            if setup.get(STRIP_CHR, True):
                chr1, chr2 = strip_chr(chr_string=chr1), strip_chr(chr_string=chr2)
            if svtype in ["DEL", "INS"]:
                if "INSLEN" in record.INFO:
                    extra[SVLEN] = int(record.INFO["INSLEN"])
                strand1 = Strand.FORWARD
                strand2 = Strand.REVERSE
            elif svtype == "DUP":
                strand1 = Strand.REVERSE
                strand2 = Strand.FORWARD
            elif svtype in ["INV", "BND"]:
                strand1, strand2 = get_strands_from_CT_vcf_string(ct_string=ct_string)
            else:
                raise Exception("SVTYPE parsing issue")
            if svtype != "INS":
                assert (strand1, strand2) == get_strands_from_CT_vcf_string(ct_string=ct_string)
            pos1 = Position(chromosome=chr1, coordinate=coord1, strand=strand1)
            pos2 = Position(chromosome=chr2, coordinate=coord2, strand=strand2)
            extra = clear_duplicated_entries_extra(extra=extra)
            na = Adjacency(position1=pos1, position2=pos2, extra=extra)
            result[svid].append(na)
        else:
            raise Exception("Unknown SV type \"{svtype}\" for Delly VCF input".format(svtype=svtype))
    result = update_nas_ids(nas_by_ids_defaultdict=result, setup=setup)
    return result.values()


def get_strands_from_CT_vcf_string(ct_string):
    strand1 = Strand.FORWARD if ct_string.startswith("3") else Strand.REVERSE
    strand2 = Strand.FORWARD if ct_string.endswith("3") else Strand.REVERSE
    return strand1, strand2


def delly_vcf_to_nas_stream(source, dest, setup=None, extra=None):
    if setup is None:
        setup = {}
    adjacencies = generate_nas_from_delly_vcf_records(source=source, setup=setup)
    write_adjacencies_to_destination(destination=dest, adjacencies=adjacencies, sort_adjacencies=False, extra=extra)


def generate_nas_from_delly_vcf_records(source, setup=None):
    vcf_reader = vcf.Reader(source)
    for record in vcf_reader:
        extra = deepcopy(record.INFO)
        svtype = record.INFO["SVTYPE"]
        svid = str(record.ID)
        extra[SVTYPE] = svtype
        extra[EXTERNAL_NA_ID] = svid
        if svtype in ["DEL", "DUP", "INV", "INS", "BND"]:
            chr1 = record.CHROM
            coord1 = int(record.POS)
            chr2 = record.INFO["CHR2"]
            coord2 = int(record.INFO["END"])
            ct_string = record.INFO["CT"]
            if setup.get(STRIP_CHR, True):
                chr1, chr2 = strip_chr(chr_string=chr1), strip_chr(chr_string=chr2)
            if svtype in ["DEL", "INS"]:
                if "INSLEN" in record.INFO:
                    extra[SVLEN] = int(record.INFO["INSLEN"])
                strand1 = Strand.FORWARD
                strand2 = Strand.REVERSE
            elif svtype == "DUP":
                strand1 = Strand.REVERSE
                strand2 = Strand.FORWARD
            elif svtype in ["INV", "BND"]:
                strand1, strand2 = get_strands_from_CT_vcf_string(ct_string=ct_string)
            else:
                raise Exception("SVTYPE parsing issue")
            if svtype != "INS":
                assert (strand1, strand2) == get_strands_from_CT_vcf_string(ct_string=ct_string)
            pos1 = Position(chromosome=chr1, coordinate=coord1, strand=strand1)
            pos2 = Position(chromosome=chr2, coordinate=coord2, strand=strand2)
            extra = clear_duplicated_entries_extra(extra=extra)
            na = Adjacency(position1=pos1, position2=pos2, extra=extra)
            if setup.get(APPEND_ID, True):
                if not na.extra[EXTERNAL_NA_ID].endswith(setup.get(ID_SUFFIX, "")):
                    na.extra[EXTERNAL_NA_ID] += "_{suffix}".format(suffix=setup.get(ID_SUFFIX, ""))
            yield na
        else:
            raise Exception("Unknown SV type \"{svtype}\" for Delly VCF input".format(svtype=svtype))


def get_nas_from_pbsv_vcf_records(pbsv_vcf_records, setup=None, sample=None):
    if setup is None:
        setup = {}
    result = defaultdict(list)
    processed_ids = set()
    records_by_ids = get_vcf_records_by_ids(vcf_records=pbsv_vcf_records)
    for record in records_by_ids.values():
        extra = deepcopy(record.INFO)
        svid = str(record.ID)
        if svid in processed_ids:
            continue
        svtype = record.INFO["SVTYPE"]
        extra[SVTYPE] = svtype
        extra[EXTERNAL_NA_ID] = svid
        if svtype in ["DEL", "INS", "INV", "BND"]:
            for vcf_sample in record.samples:
                if sample is None or vcf_sample.sample == sample:
                    support, total = 0, 0
                    l, r = vcf_sample.gt_alleles
                    if l in ['1', '.']:
                        support += vcf_sample.data.AD[0]
                    if r in ['1', '.']:
                        support += vcf_sample.data.AD[1]
                    total = vcf_sample.data.DP
                    extra["re"] = support
                    extra["rcnt"] = total
                    break
            if svtype in ["DEL", "INS"]:
                chr1 = str(record.CHROM)
                coord1 = int(record.POS)
                if setup.get(STRIP_CHR, True):
                    chr1 = strip_chr(chr_string=chr1)
                chr2 = chr1
                strand1 = Strand.FORWARD
                strand2 = Strand.REVERSE
                coord2 = record.INFO["END"]
                if svtype == "INS" and coord1 == coord2:
                    coord2 += 1
                if svtype == "INS":
                    assert len(record.INFO["SVLEN"]) == 1
                    extra[SVLEN] = int(record.INFO["SVLEN"][0])
                pos1 = Position(chromosome=chr1, coordinate=coord1, strand=strand1)
                pos2 = Position(chromosome=chr2, coordinate=coord2, strand=strand2)
                extra = clear_duplicated_entries_extra(extra=extra)
                na = Adjacency(position1=pos1, position2=pos2, extra=extra)
                result[svid].append(na)
                processed_ids.add(svid)
            elif svtype == "INV":
                chr1 = str(record.CHROM)
                if setup.get(STRIP_CHR, True):
                    chr1 = strip_chr(chr_string=chr1)
                coord11, coord12 = get_inv_na1_coords_from_vcf_record(vcf_record=record)
                coord21, coord22 = get_inv_na2_coords_from_vcf_record(vcf_record=record)
                strand11, strand12 = get_inv_na1_strands_from_vcf_record(vcf_record=record)
                strand21, strand22 = get_inv_na2_strands_from_vcf_record(vcf_record=record)
                pos11 = Position(chromosome=chr1, coordinate=coord11, strand=strand11)
                pos12 = Position(chromosome=chr1, coordinate=coord12, strand=strand12)
                pos21 = Position(chromosome=chr1, coordinate=coord21, strand=strand21)
                pos22 = Position(chromosome=chr1, coordinate=coord22, strand=strand22)
                extra1 = deepcopy(extra)
                extra2 = deepcopy(extra)
                extra1[EXTERNAL_NA_ID] = extra1[EXTERNAL_NA_ID] + "_1"
                extra2[EXTERNAL_NA_ID] = extra2[EXTERNAL_NA_ID] + "_2"
                extra1 = clear_duplicated_entries_extra(extra=extra1)
                extra2 = clear_duplicated_entries_extra(extra=extra2)
                na1 = Adjacency(position1=pos11, position2=pos12, extra=extra1)
                na2 = Adjacency(position1=pos21, position2=pos22, extra=extra2)
                result[svid].append(na1)
                result[svid].append(na2)
                processed_ids.add(svid)
            else:
                assert len(record.INFO["MATEID"]) == 1
                mate_vcf_id = record.INFO["MATEID"][0]
                if mate_vcf_id not in records_by_ids:
                    continue
                mate_record = records_by_ids[mate_vcf_id]
                extra.update(mate_record.INFO)
                chr1 = str(record.CHROM)
                coord1 = int(record.POS)
                chr2 = str(mate_record.CHROM)
                coord2 = int(mate_record.POS)
                strand1 = get_strand_from_alt_breakend(vcf_recrod=mate_record)
                strand2 = get_strand_from_alt_breakend(vcf_recrod=record)
                if setup.get(STRIP_CHR, True):
                    chr1, chr2 = strip_chr(chr_string=chr1), strip_chr(chr_string=chr2)
                pos1 = Position(chromosome=chr1, coordinate=coord1, strand=strand1)
                pos2 = Position(chromosome=chr2, coordinate=coord2, strand=strand2)
                extra[EXTERNAL_NA_ID2] = mate_vcf_id
                extra = clear_duplicated_entries_extra(extra=extra)
                na = Adjacency(position1=pos1, position2=pos2, extra=extra)
                result[svid].append(na)
                processed_ids.add(svid)
                processed_ids.add(mate_vcf_id)
        else:
            raise Exception("Unknonw SV type \"{svtype}\" for PBSV VCF input".format(svtype=svtype))
    result = update_nas_ids(nas_by_ids_defaultdict=result, setup=setup)
    return result.values()


GUNDEM_SAMPLE_NAME = "patient"
GUNDEM_CHR1 = "chr1"
GUNDEM_CHR2 = "chr2"
GUNDEM_COORDINATE1 = "pos1"
GUNDEM_COORDINATE2 = "pos2"
GUNDEM_STRAND1 = "str1"
GUNDEM_STRAND2 = "str2"
GUNDEM_SAMPLES_LIST = "samples"
GUNDEM_SAMPLES_COUNTS = "tumour-counts"
GUNDEM_PATIENT = "patient"

GUNDEM_PER_SAMPLE_SUPPORT = "per_sample_support"


def gundem2015_flip_strand(strand_string):
    if strand_string == "-":
        return "+"
    elif strand_string == "+":
        return "-"
    raise ValueError("{} is not a valid strand string".format(strand_string))


def get_nas_from_gundem2015_file(file_name, separator="\t", flip_second_strand=True):
    with open(file_name, "rt") as source:
        return get_nas_from_gundem2015_source(source=source, separator=separator, flip_second_strand=flip_second_strand)


def get_nas_from_gundem2015_source(source, setup=None, separator="\t", flip_second_strand=True):
    if setup is None:
        setup = {}
    result = defaultdict(list)
    reader = csv.DictReader(source, delimiter=separator)
    for cnt, row in enumerate(reader):
        patient_name = row[GUNDEM_PATIENT]
        sample_list = row[GUNDEM_SAMPLES_LIST].split("/")
        sample_list = [patient_name + "-" + sample_name for sample_name in sample_list]
        sample_cnts = list(map(int, row[GUNDEM_SAMPLES_COUNTS].split("/")))
        per_sample_counts = {sample: cnt for sample, cnt in zip(sample_list, sample_cnts)}
        chromosome1, chromosome2 = row[GUNDEM_CHR1], row[GUNDEM_CHR2]
        if setup.get(STRIP_CHR, True):
            chromosome1, chromosome2 = strip_chr(chr_string=chromosome1), strip_chr(chr_string=chromosome2)
        position1 = Position(chromosome=chromosome1, coordinate=int(row[GUNDEM_COORDINATE1]), strand=Strand.from_pm_string(row[GUNDEM_STRAND1]))
        strand2_string = gundem2015_flip_strand(row[GUNDEM_STRAND2]) if flip_second_strand else row[GUNDEM_STRAND2]
        position2 = Position(chromosome=chromosome2, coordinate=int(row[GUNDEM_COORDINATE2]), strand=Strand.from_pm_string(strand2_string))
        aid = str(cnt)

        extra = {
            EXTERNAL_NA_ID: aid,
            GUNDEM_PER_SAMPLE_SUPPORT: per_sample_counts,
        }
        novel_adjacency = Adjacency(position1=position1, position2=position2, adjacency_type=AdjacencyType.NOVEL, extra=extra)
        result[aid].append(novel_adjacency)

    result = update_nas_ids(nas_by_ids_defaultdict=result, setup=setup)
    return result.values()


REMIXT_PREDICTION_ID = "prediction_id"
REMIXT_CHROMOSOME_1 = "chromosome_1"
REMIXT_POSITION_1 = "position_1"
REMIXT_STRAND_1 = "strand_1"
REMIXT_CHROMOSOME_2 = "chromosome_2"
REMIXT_POSITION_2 = "position_2"
REMIXT_STRAND_2 = "strand_2"
REMIXT_CN_1 = "cn_1"
REMIXT_CN_2 = "cn_2"


def get_nas_from_remixt_file(file_name, setup=None, separator="\t", clone_ids=None, remixt_na_correction=True, skip_absent=False):
    with open(file_name, "rt") as source:
        return get_nas_from_remixt_source(source=source, setup=setup, separator=separator, clone_ids=clone_ids, remixt_na_correction=remixt_na_correction, skip_absent=skip_absent)


def get_nas_from_remixt_source(source, setup=None, separator="\t", clone_ids=None, remixt_na_correction=True, skip_absent=False):
    if clone_ids is None:
        clone_ids = ["1", "2"]
    if setup is None:
        setup = {}
    result = defaultdict(list)
    reader = csv.DictReader(source, delimiter=separator)
    for line in reader:
        naid = line[REMIXT_PREDICTION_ID]
        chr1_name = line[REMIXT_CHROMOSOME_1]
        chr1_coordinate = int(line[REMIXT_POSITION_1])
        chr1_strand = Strand.from_pm_string(line[REMIXT_STRAND_1])
        chr2_name = line[REMIXT_CHROMOSOME_2]
        chr2_coordinate = int(line[REMIXT_POSITION_2])
        chr2_strand = Strand.from_pm_string(line[REMIXT_STRAND_2])
        clone1_cn = int(line[REMIXT_CN_1])
        clone2_cn = int(line[REMIXT_CN_2])
        cn_data = {}
        skip = True
        for clone_cn, clone_name in zip([clone1_cn, clone2_cn], ["1", "2"]):
            if clone_name in clone_ids:
                cn_data[clone_name] = {str(Phasing.AA): clone2_cn}
                skip &= clone2_cn == 0
        skip = skip and skip_absent
        if skip:
            continue
        if remixt_na_correction and chr1_strand == Strand.FORWARD:
            chr1_coordinate -= 1
        if remixt_na_correction and chr2_strand == Strand.FORWARD:
            chr2_coordinate -= 1
        position1 = Position(chromosome=chr1_name, coordinate=chr1_coordinate, strand=chr1_strand)
        position2 = Position(chromosome=chr2_name, coordinate=chr2_coordinate, strand=chr2_strand)
        novel_adjacency = Adjacency(position1=position1, position2=position2,
                                    adjacency_type=AdjacencyType.NOVEL, extra={EXTERNAL_NA_ID: naid, COPY_NUMBER: cn_data})
        aid = novel_adjacency.stable_id_non_phased
        result[aid].append(novel_adjacency)
    result = update_nas_ids(nas_by_ids_defaultdict=result, setup=setup)
    return result.values()


def build_setup(args):
    result = {
        STRIP_CHR: args.strip_chr,
        APPEND_ID: args.append_id_suffix,
        EXTRA_FIELDS: args.o_extra_fields.split(","),
        ID_SUFFIX: args.id_suffix,
    }
    return result


def get_chrs_regions_string_list_from_file(file_name):
    with open(file_name, "rt") as source:
        return get_chrs_regions_string_lists_from_source(source=source)


def get_chrs_regions_string_lists_from_source(source):
    result = []
    for line in source:
        line = line.strip()
        if len(line) == 0 or line.startswith("#"):
            continue
        result.append(line)
    return result


def parse_segment_chr_region(string):
    if ":" not in string and "\t" not in string:
        return Segment.from_chromosome_coordinates(chromosome=string.strip(), start=0, end=3000000000)
    if ":" in string:
        chromosome, start_end = string.split(":")
        start, end = start_end.split("-")
        start, end = int(start), int(end)
        return Segment.from_chromosome_coordinates(chromosome=chromosome.lower(), start=start, end=end)
    chromosome, start, end = string.split("\t")
    start, end = int(start), int(end)
    return Segment.from_chromosome_coordinates(chromosome=chromosome.lower(), start=start, end=end)




from copy import deepcopy

import networkx as nx

from rck.core.structures import propagate_haplotype_segment_to_positions, propagate_phasing_adjacency_to_positions
from rck.core.structures import SegmentCopyNumberProfile, AdjacencyCopyNumberProfile, StructureProfile, Segment, Adjacency, AdjacencyType, HAPLOTYPE, Haplotype, Phasing, PHASING

COPY_NUMBER = "copy_number"


def edge_tuple_based_on_flag(u, v, attr, data):
    if data:
        return u, v, attr
    else:
        return u, v


def node_tuple_based_on_flag(n, attr, data):
    if data:
        return n, attr
    return n


class IntervalAdjacencyGraph(object):
    def __init__(self, segments=None, adjacencies=None):
        segments = segments if segments is not None else []
        adjacencies = adjacencies if adjacencies is not None else []
        self.segments = segments
        self.adjacencies = adjacencies
        self.graph = nx.MultiGraph()

    @property
    def has_proper_topology(self):
        for node, n_data in self.nodes(data=True):
            s_edges = list(self.segment_edges(nbunch=node, data=True))
            if len(s_edges) != 1:    # every vertex (i.e., segment's extremity, can have at most 1 segment edge incident to it)
                return False
            if s_edges[0][0] == s_edges[0][1]:  # segments can not be represented in a form of self-loops
                return False
        return True

    @classmethod
    def get_segment_object_from_segment(cls, segment, copy=True):
        return segment.get_non_hap_copy()

    @classmethod
    def get_adjacency_object_from_adjacency(cls, adjacency, copy=True):
        return adjacency.get_non_phased_copy()

    def add_segment_edge(self, segment, sort=True, copy_segment=True):
        u, v = self.get_edge_vertices_pair_from_segment(segment=segment, sort=sort)
        if self.has_segment_edge_by_edge(edge=(u, v), sort=True):
            return
        obj = self.get_segment_object_from_segment(segment=segment, copy=copy_segment)
        self.graph.add_edge(u, v, object=obj)

    def add_adjacency_edge(self, adjacency, sort=True, copy_adjacency=True):
        u, v = self.get_edge_vertices_pair_from_adjacency(adjacency=adjacency, sort=sort)
        if u not in self.graph:
            raise ValueError()
        if v not in self.graph:
            raise ValueError()
        obj = self.get_adjacency_object_from_adjacency(adjacency=adjacency, copy=copy_adjacency)
        for adj_edge_w_data in self.adjacency_edges(nbunch=u):
            adj_u, adj_v, adj_data = adj_edge_w_data
            if {adj_u, adj_v} == {u, v} and adj_data["object"].adjacency_type == obj.adjacency_type:
                return
        # if self.has_adjacency_edge(edge=(u, v), sort=True):   # can not have parallel adjacency edges. Only possible parallel edges are pairs of segment/adjacency ones
        #     return
        self.graph.add_edge(u, v, object=obj)

    @staticmethod
    def get_edge_vertices_pair_from_adjacency(adjacency, sort=True):
        u, v = adjacency.position1.get_non_hap_copy(), adjacency.position2.get_non_hap_copy()
        if sort:
            u, v = tuple(sorted([u, v]))
        return u, v

    @staticmethod
    def get_edge_vertices_pair_from_segment(segment, sort=True):
        u, v = segment.start_position.get_non_hap_copy(), segment.end_position.get_non_hap_copy()
        if sort and segment.is_reversed:
            u, v = v, u
        return u, v

    def build_graph(self):
        for s in self.segments:
            self.add_segment_edge(segment=s)
        for a in self.adjacencies:
            self.add_adjacency_edge(adjacency=a)

    def nodes(self, data=True):
        for n, attr in self.graph.nodes(data=True):
            yield node_tuple_based_on_flag(n=n, attr=attr, data=data)

    def edges(self, data=True, nbunch=None):
        for value in self.graph.edges(nbunch=nbunch, data=data):
            yield value

    def segment_edges(self, data=True, nbunch=None, sort=True):
        for u, v, attr in self.edges(nbunch=nbunch, data=True):
            if isinstance(attr["object"], Segment):
                if sort:
                    u, v = tuple(sorted([u, v]))
                yield edge_tuple_based_on_flag(u, v, attr, data)

    def adjacency_edges(self, data=True, nbunch=None, sort=True):
        for u, v, attr in self.edges(nbunch=nbunch, data=True):
            if isinstance(attr["object"], Adjacency):
                if sort:
                    u, v = tuple(sorted([u, v]))
                yield edge_tuple_based_on_flag(u=u, v=v, attr=attr, data=data)

    def ref_adjacency_edges(self, data=True, nbunch=None, sort=True):
        for u, v, attr in self.adjacency_edges(data=True, nbunch=nbunch, sort=sort):
            if attr["object"].adjacency_type == AdjacencyType.REFERENCE:
                yield edge_tuple_based_on_flag(u=u, v=v, attr=attr, data=data)

    def nov_adjacency_edges(self, data=True, nbunch=None, sort=True):
        for u, v, attr in self.adjacency_edges(data=True, nbunch=nbunch, sort=sort):
            if attr["object"].adjacency_type == AdjacencyType.NOVEL:
                yield edge_tuple_based_on_flag(u=u, v=v, attr=attr, data=data)

    def get_segment_edge(self, node, data=True, sort=True):
        segment_edges = list(self.segment_edges(data=True, nbunch=node, sort=sort))
        if len(segment_edges) != 1:
            raise ValueError("Vertex {vertex} does no either 0 or more than 1 segment edge incident to it".format(vertex=str(node)))
        u = segment_edges[0][0]
        v = segment_edges[0][1]
        attr = segment_edges[0][2]
        return edge_tuple_based_on_flag(u=u, v=v, attr=attr, data=data)

    def has_segment_edge_by_edge(self, edge, sort=True):
        u, v = edge
        if sort:
            u, v = tuple(sorted([u, v]))
        if u not in self.graph or v not in self.graph:
            return False
        try:
            _, __, u_s_edge_data = self.get_segment_edge(node=u, data=True, sort=True)
            _, __, v_s_edge_data = self.get_segment_edge(node=v, data=True, sort=True)
            if u_s_edge_data["object"] != v_s_edge_data["object"]:
                return False
            return True
        except ValueError:
            return False

    def has_adjacency_edge_by_edge(self, edge, sort=True):
        u, v, = edge
        if sort:
            u, v = tuple(sorted([u, v]))
        if u not in self.graph or v not in self.graph:
            return False
        try:
            exists = False
            for u_, u__, u_a_edge_data in self.adjacency_edges(nbunch=u, data=True, sort=True):
                for v_, v__, v_a_edge_data in self.adjacency_edges(nbunch=v, data=True, sort=True):
                    if u_a_edge_data["object"] == v_a_edge_data["object"]:
                        exists = True
            return exists
        except ValueError:
            return False

    @property
    def complies_with_is(self):
        for node in self.nodes(data=False):
            n_as = list(self.nov_adjacency_edges(nbunch=node))
            if len(n_as) > 1:
                return False
        return True

    @property
    def compatible_with_hsis(self):
        for node in self.nodes(data=False):
            n_as = list(self.nov_adjacency_edges(nbunch=node))
            if len(n_as) > 2:
                return False
        return True

    @property
    def complies_with_extremity_exclusivity(self):
        return len(self.violations_of_extremity_exclusivity()) == 0

    def violations_of_extremity_exclusivity(self):
        violations = []
        for node in self.nodes(data=False):
            nas = list(self.nov_adjacency_edges(nbunch=node))
            if len(nas) > 1:
                violations.append(node)
        return violations

    def adjacency_edges_connected_components_subgraphs(self, ref=True, nov=True, copy=True):
        adjacency_edge_only_iag = self.__class__()
        if ref:
            for u, v, data in self.ref_adjacency_edges(data=True, sort=True):
                adjacency_edge_only_iag.graph.add_edge(u, v, **data)
        if nov:
            for u, v, data in self.nov_adjacency_edges(data=True, sort=True):
                adjacency_edge_only_iag.graph.add_edge(u, v, **data)
        for adj_iag_cc in nx.connected_component_subgraphs(G=adjacency_edge_only_iag.graph, copy=copy):
            iag = self.__class__()
            iag.graph = adj_iag_cc
            yield iag

    def ref_adjacency_edges_connected_components_subgraphs(self, copy=True):
        for iag in self.adjacency_edges_connected_components_subgraphs(ref=True, nov=False, copy=copy):
            yield iag

    def nov_adjacency_edges_connected_components_subgraphs(self, copy=True):
        for iag in self.adjacency_edges_connected_components_subgraphs(ref=False, nov=True, copy=copy):
            yield iag

    def get_set_self_segment_edges(self, sort=True):
        result = set()
        for u, v in self.segment_edges(data=False, sort=sort):
            result.add((u, v))
        return result

    def get_set_self_adjacency_edges(self, sort=True):
        result = set()
        for u, v in self.adjacency_edges(data=False, sort=sort):
            result.add((u, v))
        return result

    def intersection_on_segment_edges(self, segments):
        """iterates over `segments` once"""
        other_edges = set()
        for s in segments:
            u, v = self.get_edge_vertices_pair_from_segment(segment=s, sort=True)
            other_edges.add((u, v))
        return self.get_set_self_segment_edges(sort=True).intersection(other_edges)

    def intersection_on_adjacency_edges(self, adjacencies):
        """iterates over `adjacencies` once"""
        other_edges = set()
        for a in adjacencies:
            u, v = self.get_edge_vertices_pair_from_adjacency(adjacency=a, sort=True)
            other_edges.add((u, v))
        return self.get_set_self_adjacency_edges(sort=True).intersection(other_edges)

    def relative_complements_on_segment_edges(self, segments):
        """iterates over `segments` once"""
        self_edges = self.get_set_self_segment_edges(sort=True)
        other_edges = set()
        for s in segments:
            u, v = self.get_edge_vertices_pair_from_segment(segment=s, sort=True)
            other_edges.add((u, v))
        intersection = self_edges.intersection(other_edges)
        self_rel_compl = self_edges - intersection
        other_rel_compl = other_edges - intersection
        return self_rel_compl, other_rel_compl

    def relative_complements_on_adjacency_edges(self, adjacencies):
        """iterates over `adjacencies` once"""
        self_edges = self.get_set_self_adjacency_edges(sort=True)
        other_edges = set()
        for a in adjacencies:
            u, v = self.get_edge_vertices_pair_from_adjacency(adjacency=a, sort=True)
            other_edges.add((u, v))
        intersection = self_edges.intersection(other_edges)
        self_rel_compl = self_edges - intersection
        other_rel_compl = other_edges - intersection
        return self_rel_compl, other_rel_compl

    def topology_matches_for_genome(self, genome):
        self_rel_compl_segments, genome_rel_compl_segments = self.relative_complements_on_segment_edges(segments=genome.iter_segments())
        if len(self_rel_compl_segments) != 0 or len(genome_rel_compl_segments) != 0:
            return False
        self_rel_compl_adjs, genome_rel_compl_adjs = self.relative_complements_on_adjacency_edges(adjacencies=genome.iter_adjacencies())
        if len(self_rel_compl_adjs) != 0 or len(genome_rel_compl_adjs) != 0:
            return False
        return True

    def topology_allows_for_genome(self, genome):
        self_rel_compl_segments, genome_rel_compl_segments = self.relative_complements_on_segment_edges(segments=genome.iter_segments())
        if len(genome_rel_compl_segments) != 0:
            return False
        self_rel_compl_adjs, genome_rel_compl_adjs = self.relative_complements_on_adjacency_edges(adjacencies=genome.iter_adjacencies())
        if len(genome_rel_compl_adjs) != 0:
            return False
        return True

    def represents_given_genome(self, genome):
        if not self.is_copy_number_aware:
            raise ValueError()
        scn_profile = SegmentCopyNumberProfile.from_genome(genome=genome)
        if not self.matches_segment_copy_number_profile(scn_profile=scn_profile):
            return False
        acn_profile = AdjacencyCopyNumberProfile.from_genome(genome=genome)
        if not self.matches_adjacency_copy_number_profile(acn_profile=acn_profile):
            return False
        return True

    @property
    def is_copy_number_aware(self):
        for u, v, data in self.edges(data=True):
            if COPY_NUMBER not in data:
                return False
        return True

    @property
    def represents_a_genome(self):
        if not self.is_copy_number_aware:
            return False
        for node in self.graph:
            if not self.has_non_negative_imbalance(node=node):
                return False
        return True

    def node_imbalance(self, node):
        u, v, data = self.get_segment_edge(node=node, data=True, sort=True)
        if COPY_NUMBER not in data:
            raise ValueError("Segment edge {p1}-{p2} does not have a copy numbe associated with it".format(p1=str(u), p2=str(u)))
        segment_copy_number = data[COPY_NUMBER]
        imbalance = segment_copy_number
        for u, v, data in self.adjacency_edges(data=True, nbunch=node, sort=True):
            if COPY_NUMBER not in data:
                raise ValueError()
            adj_copy_number = data[COPY_NUMBER]
            adj_copy_number_multiplier = 2 if u == v else 1
            imbalance -= (adj_copy_number_multiplier * adj_copy_number)
        return imbalance

    def has_non_negative_imbalance(self, node):
        return self.node_imbalance(node=node) >= 0

    def has_positive_imbalance(self, node):
        return self.node_imbalance(node=node) > 0

    def has_zero_imbalance(self, node):
        return self.node_imbalance(node=node) == 0

    def is_telomere(self, node):
        return self.has_positive_imbalance(node=node)

    def is_non_telomere(self, node):
        return self.has_zero_imbalance(node=node)

    def matches_segment_copy_number_profile(self, scn_profile):
        sid_keys_set_from_profile = set(scn_profile.sid_keys())
        observed_sid_keys_set_from_graph = set()
        for u, v, data in self.segment_edges(data=True, sort=True):
            segment_obj = data["object"]
            cn = data[COPY_NUMBER]
            sid = segment_obj.stable_id_non_hap
            if scn_profile.get_combined_cn(sid=sid, default=0) != cn:
                return False
            observed_sid_keys_set_from_graph.add(sid)
        non_observed_sid_keys_set = sid_keys_set_from_profile - observed_sid_keys_set_from_graph
        for non_observed_sid in non_observed_sid_keys_set:
            if scn_profile.get_combined_cn(sid=non_observed_sid, default=0) != 0:
                return False
        return True

    def remove_segment_edge(self, u, v, silent=True):
        has_edge = self.has_segment_edge_by_edge(edge=(u, v), sort=True)
        if not has_edge:
            if not silent:
                raise ValueError("Attempted to remove segment edge {p1}-{p2}, but it is not present in the graph".format(p1=str(u), p2=str(v)))
            return
        edges_to_remove = []
        for u, v, key, data in self.graph.edges(nbunch=u, data=True, keys=True):
            obj = data["object"]
            if isinstance(obj, Segment):
                edges_to_remove.append((u, v, key))
        for u, v, key in edges_to_remove:
            self.graph.remove_edge(u=u, v=v, key=key)

    def remove_adjacency_edge(self, u, v, silent=True):
        has_edge = self.has_adjacency_edge_by_edge(edge=(u, v), sort=True)
        if not has_edge:
            if not silent:
                raise ValueError("Attempted to remove adjacency edge {p1}-{p2}, but it is not present in the graph".format(p1=str(u), p2=str(v)))
            return
        edges_to_remove = []
        for u_, u__, u_key, u_data in self.graph.edges(nbunch=u, data=True, keys=True):
            if not isinstance(u_data["object"], Adjacency):
                continue
            for v_, v__, v_key, v_data in self.graph.edges(nbunch=v, data=True, keys=True):
                if not isinstance(v_data["object"], Adjacency):
                    continue
                if u_data["object"] == v_data["object"] and u_data["object"].adjacency_type == v_data["object"].adjacency_type:
                    assert v_key == u_key
                    edges_to_remove.append((u, v, v_key))
        for u, v, key in edges_to_remove:
            self.graph.remove_edge(u=u, v=v, key=key)

    def remove_edges_with_zero_cn(self, check_cn_awareness=False, clear_nodes_after=True):
        self.remove_adjacency_edges_with_zero_cn(check_cn_awareness=check_cn_awareness, clear_nodes_after=clear_nodes_after)
        self.remove_segment_edges_with_zero_cn(check_cn_awareness=check_cn_awareness, clear_nodes_after=clear_nodes_after)

    def remove_segment_edges_with_zero_cn(self, check_cn_awareness=False, clear_nodes_after=True):
        if check_cn_awareness and not self.is_copy_number_aware:
            raise ValueError("Graph does not have copy number information associated with (at least some) its edges.")
        segment_edges_to_remove = []
        for u, v, data in self.segment_edges(data=True, sort=True):
            cn = data[COPY_NUMBER]
            if cn == 0:
                segment_edges_to_remove.append((u, v))
        for u, v in segment_edges_to_remove:
            self.remove_segment_edge(u=u, v=v)
            if clear_nodes_after:
                if self.graph.degree[u] == 0:
                    self.graph.remove_node(n=u)
                if self.graph.degree[v] == 0:
                    self.graph.remove_node(n=v)

    def remove_adjacency_edges_with_zero_cn(self, check_cn_awareness=False, clear_nodes_after=True):
        self.remove_ref_adjacency_edges_with_zero_cn(check_cn_awareness=check_cn_awareness, clear_nodes_after=clear_nodes_after)
        self.remove_nov_adjacency_edges_with_zero_cn(check_cn_awareness=check_cn_awareness, clear_nodes_after=clear_nodes_after)

    def remove_ref_adjacency_edges_with_zero_cn(self, check_cn_awareness=False, clear_nodes_after=True):
        if check_cn_awareness and not self.is_copy_number_aware:
            raise ValueError("Graph does not have copy number data associated with its edges")
        adjacency_edges_to_remove = []
        for u, v, data in self.ref_adjacency_edges(data=True, sort=True):
            cn = data[COPY_NUMBER]
            if cn == 0:
                adjacency_edges_to_remove.append((u, v))
                # self.remove_adjacency_edge(u=u, v=v)
        for u, v in adjacency_edges_to_remove:
            self.remove_adjacency_edge(u=u, v=v)
            if clear_nodes_after:
                if self.graph.degree[u] == 0:
                    self.graph.remove_node(n=u)
                if self.graph.degree[v] == 0:
                    self.graph.remove_node(n=v)

    def remove_nov_adjacency_edges_with_zero_cn(self, check_cn_awareness=False, clear_nodes_after=True):
        if check_cn_awareness and not self.is_copy_number_aware:
            raise ValueError()
        adjacency_edges_to_remove = []
        for u, v, data in self.nov_adjacency_edges(data=True, sort=True):
            cn = data[COPY_NUMBER]
            if cn == 0:
                adjacency_edges_to_remove.append((u, v))
                # self.remove_adjacency_edge(u=u, v=v)
        for u, v, in adjacency_edges_to_remove:
            self.remove_adjacency_edge(u=u, v=v)
            if clear_nodes_after:
                if self.graph.degree[u] == 0:
                    self.graph.remove_node(n=u)
                if self.graph.degree[v] == 0:
                    self.graph.remove_node(n=v)

    def matches_adjacency_copy_number_profile(self, acn_profile):
        aid_keys_set_from_profile = set(acn_profile.aid_keys())
        observed_aid_keys_from_graph = set()
        for u, v, data in self.adjacency_edges(data=True, sort=True):
            adjacency_obj = data["object"]
            cn = data[COPY_NUMBER]
            aid = adjacency_obj.stable_id_non_phased
            if acn_profile.get_combined_cn(aid=aid, default=0) != cn:
                return False
            observed_aid_keys_from_graph.add(aid)
        non_observed_aid_keys_set = aid_keys_set_from_profile - observed_aid_keys_from_graph
        for non_observed_aid in non_observed_aid_keys_set:
            if acn_profile.get_combined_cn(aid=non_observed_aid, default=0) != 0:
                return False
        return True

    def assign_copy_numbers_from_genome(self, genome, ensure_topology=True):
        if ensure_topology and not self.topology_allows_for_genome(genome=genome):
            raise ValueError()
        self.assign_copy_numbers_from_structure_profile(structure_profile=StructureProfile.from_genome(genome=genome))

    def assign_copy_numbers_from_structure_profile(self, structure_profile):
        self.assign_copy_numbers_from_scn_profile(scn_profile=structure_profile.scn_profile)
        self.assign_copy_numbers_from_acn_profile(acn_profile=structure_profile.acn_profile)

    def assign_copy_numbers_from_scn_profile(self, scn_profile):
        for u, v, data in self.segment_edges(data=True, sort=True):
            data[COPY_NUMBER] = scn_profile.get_combined_cn(sid=data["object"].stable_id_non_hap, default=0)

    def assign_copy_numbers_from_acn_profile(self, acn_profile):
        for u, v, data in self.adjacency_edges(data=True, sort=True):
            data[COPY_NUMBER] = acn_profile.get_combined_cn(aid=data["object"].stable_id_non_phased, default=0)

    def has_adjacency_edge(self, edge, sort=True):
        try:
            self.get_adjacency_edge(edge=edge, data=False, sort=sort)
            return True
        except ValueError:
            return False

    def get_adjacency_edge(self, edge, data=True, sort=True):
        u, v = edge
        if sort:
            u, v = tuple(sorted([u, v]))
        u_adjacency_edges = list(self.adjacency_edges(data=True, nbunch=u, sort=True))
        if len(u_adjacency_edges) == 0:
            raise ValueError()
        v_adjacency_edges = list(self.adjacency_edges(data=True, nbunch=v, sort=True))
        if len(v_adjacency_edges) == 0:
            raise ValueError()
        shared_edge = None
        for uu, uv, uattr in u_adjacency_edges:
            if (uu, uv) != (u, v):
                continue
            for vu, vv, vattr in v_adjacency_edges:
                if (uu, uv) == (vu, vv) and uattr["object"] == vattr["object"]:
                    shared_edge = u, v, uattr
                    break
            if shared_edge is not None:
                break
        if shared_edge is None:
            raise ValueError()
        u, v, attr = shared_edge
        return edge_tuple_based_on_flag(u=u, v=v, attr=attr, data=data)

    def get_telomeres(self, check_cn_awareness=True, sort=True, copy=True):
        result = [t for t in self.iter_telomeres(check_cn_awareness=check_cn_awareness)]
        if copy:
            result = deepcopy(result)
        if sort:
            result = sorted(result)
        return result

    def iter_telomeres(self, check_cn_awareness=True):
        if check_cn_awareness and not self.is_copy_number_aware:
            raise ValueError()
        for node in self.nodes(data=False):
            if self.is_telomere(node=node):
                yield node


IAG = IntervalAdjacencyGraph


class HaplotypeSpecificIntervalAdjacencyGraph(IntervalAdjacencyGraph):
    def __init__(self, segments=None, adjacencies=None):
        super(HaplotypeSpecificIntervalAdjacencyGraph, self).__init__(segments=segments, adjacencies=adjacencies)

    @classmethod
    def get_segment_object_from_segment(cls, segment, copy=True):
        return deepcopy(segment) if copy else segment

    @classmethod
    def get_adjacency_object_from_adjacency(cls, adjacency, copy=True):
        return deepcopy(adjacency) if copy else adjacency

    def add_segment_edge(self, segment, sort=True, copy_segment=True):
        if not segment.is_haplotype_specific:
            raise ValueError()
        super(HaplotypeSpecificIntervalAdjacencyGraph, self).add_segment_edge(segment=segment, sort=sort)

    @classmethod
    def get_edge_vertices_pair_from_segment(cls, segment, sort=True):
        u, v = deepcopy(segment.start_position), deepcopy(segment.end_position)
        if sort and segment.is_reversed:
            u, v = v, u
        return u, v

    @classmethod
    def get_edge_vertices_pair_from_adjacency(cls, adjacency, sort=True):
        u, v = deepcopy(adjacency.position1), deepcopy(adjacency.position2)
        if sort:
            u, v = tuple(sorted([u, v]))
        return u, v

    def add_adjacency_edge(self, adjacency, sort=True, copy_adjacency=True):
        if not adjacency.is_phased:
            raise ValueError()
        super(HaplotypeSpecificIntervalAdjacencyGraph, self).add_adjacency_edge(adjacency=adjacency, sort=sort)

    def build_graph(self):
        for s in self.segments:
            self.add_segment_edge(segment=s)
        for a in self.adjacencies:
            self.add_adjacency_edge(adjacency=a)

    @property
    def complies_with_hiis(self):
        processed_nodes = set()
        for node, attr in self.nodes(data=True):
            n_as = list(self.nov_adjacency_edges(nbunch=node))
            if node in processed_nodes:
                continue
            position = node
            processed_nodes.add(position)
            p_haplotype = position.extra[HAPLOTYPE]
            hap_mates = [hap for hap in Haplotype if hap != p_haplotype]
            mate_positions = [deepcopy(position) for _ in hap_mates]
            for p, mh in zip(mate_positions, hap_mates):
                p.extra[HAPLOTYPE] = mh
                processed_nodes.add(p)
            for mp in mate_positions:
                if mp in self.graph:
                    n_as.extend(list(self.nov_adjacency_edges(nbunch=mp)))
            if len(n_as) > 1:
                return False
        return True

    @property
    def complies_with_hsis(self):
        return self.complies_with_is

    @property
    def compatible_with_hsis(self):
        return self.complies_with_hsis

    @property
    def complies_with_homologous_extremity_exclusivity(self):
        return len(self.violations_of_homologous_extremity_exclusivity()) == 0

    def violations_of_homologous_extremity_exclusivity(self):
        violations = []
        processed_nodes = set()
        for node, attr in self.nodes(data=True):
            if node in processed_nodes:
                continue
            n_as = list(self.nov_adjacency_edges(nbunch=node))
            if len(n_as) == 0:
                continue
            position = node
            processed_nodes.add(position)
            p_haplotype = position.extra[HAPLOTYPE]
            hap_mates = [hap for hap in (Haplotype.A, Haplotype.B) if hap != p_haplotype]
            mate_positions = [position.get_non_hap_copy() for _ in hap_mates]
            for p, mh in zip(mate_positions, hap_mates):
                p.extra[HAPLOTYPE] = mh
                processed_nodes.add(p)
            for mp in mate_positions:
                if mp in self.graph:
                    mate_nas = list(self.nov_adjacency_edges(nbunch=mp))
                    if len(mate_nas) > 0:
                        violations.append((node, mate_positions))
                        break
        return violations

    @property
    def complies_with_homologous_reciprocal_extremity_exclusivity(self):
        return len(self.violations_of_homologous_reciprocal_extremity_exclusivity()) == 0

    def violations_of_homologous_reciprocal_extremity_exclusivity(self):
        violators = []
        processed_nodes = set()
        for node, attr in self.nodes(data=True):
            if node in processed_nodes:
                continue
            ref_adjacencies = list(self.ref_adjacency_edges(nbunch=node))
            if len(ref_adjacencies) == 0:
                continue
            nas = list(self.nov_adjacency_edges(nbunch=node))
            if len(nas) == 0:
                continue
            for ref_adjacency in ref_adjacencies:
                u, v, data = ref_adjacency
                reciprocal_mate = u if v == node else v
                p_haplotype = reciprocal_mate.extra[HAPLOTYPE]
                hap_mates = [hap for hap in (Haplotype.A, Haplotype.B) if hap != p_haplotype]
                mate_positions = [reciprocal_mate.get_non_hap_copy() for _ in hap_mates]
                for p, mh in zip(mate_positions, hap_mates):
                    p.extra[HAPLOTYPE] = mh
                    processed_nodes.add(p)
                for mp in mate_positions:
                    if mp in self.graph:
                        if len(list(self.nov_adjacency_edges(nbunch=mp))) > 0:
                            violators.append((node, mp))
        return violators

    def matches_segment_copy_number_profile(self, scn_profile):
        sid_hap_pairs_set_from_profile = set(scn_profile.sid_hap_pairs())
        observed_sid_hap_pairs_set_from_graph = set()
        for u, v, data in self.segment_edges(data=True, sort=True):
            segment_object = data["object"]
            cn = data[COPY_NUMBER]
            hap = segment_object.haplotype
            sid = segment_object.stable_id_non_hap
            if scn_profile.get_cn(sid=sid, haplotype=hap, default=0) != cn:
                return False
            observed_sid_hap_pairs_set_from_graph.add((sid, hap))
        non_observed_sid_hap_pairs_set = sid_hap_pairs_set_from_profile - observed_sid_hap_pairs_set_from_graph
        for sid, hap in non_observed_sid_hap_pairs_set:
            if scn_profile.get_cn(sid=sid, hap=hap, default=0) != 0:
                return False
        return True

    def matches_adjacency_copy_number_profile(self, acn_profile):
        aid_phase_pairs_set_from_profile = set(acn_profile.aid_phase_pairs())
        observed_aid_phase_pairs_set_from_graph = set()
        for u, v, data in self.adjacency_edges(data=True, sort=True):
            adj_object = data["object"]
            cn = data[COPY_NUMBER]
            aid = adj_object.stable_id_non_phased
            phasing = adj_object.stable_phasing
            if acn_profile.get_cn(aid=aid, phasing=phasing, default=0) != cn:
                return False
            observed_aid_phase_pairs_set_from_graph.add((aid, phasing))
        non_observed_aid_phase_pairs_set = aid_phase_pairs_set_from_profile - observed_aid_phase_pairs_set_from_graph
        for aid, phasing in non_observed_aid_phase_pairs_set:
            if acn_profile.get_cn(aid=aid, phasing=phasing, default=0) != 0:
                return False
        return True

    def assign_copy_numbers_from_scn_profile(self, scn_profile):
        for u, v, data in self.segment_edges(data=True, sort=True):
            sid = data["object"].stable_id_non_hap
            hap = data["object"].haplotype
            data[COPY_NUMBER] = scn_profile.get_cn(sid=sid, haplotype=hap, default=0)

    def assign_copy_numbers_from_acn_profile(self, acn_profile):
        for u, v, data in self.adjacency_edges(data=True, sort=True):
            adj = data["object"]
            aid = adj.stable_id_non_phased
            phasing = adj.stable_phasing
            data[COPY_NUMBER] = acn_profile.get_cn(aid=aid, phasing=phasing, default=0)


def construct_iag(ref_genome, mut_genomes):
    ref_adjacencies = {}
    for adj in ref_genome.iter_adjacencies(adjacency_type=AdjacencyType.REFERENCE):
        aid = adj.stable_id_non_phased
        if aid not in ref_adjacencies:
            ref_adjacencies[aid] = adj
    nov_adjacencies = {}
    for mut_genome in mut_genomes:
        for adj in mut_genome.iter_adjacencies(adjacency_type=AdjacencyType.NOVEL):
            aid = adj.stable_id_non_phased
            if aid not in ref_adjacencies and aid not in nov_adjacencies:
                nov_adjacencies[aid] = adj
    segments = {}
    for s in ref_genome.iter_segments():
        sid = s.stable_id_non_hap
        if sid not in segments:
            segments[sid] = s
    result = IntervalAdjacencyGraph()
    for segment in segments.values():
        result.add_segment_edge(segment=segment, sort=True, copy_segment=True)
    for adj in ref_adjacencies.values():
        result.add_adjacency_edge(adjacency=adj, sort=True, copy_adjacency=True)
    for adj in nov_adjacencies.values():
        result.add_adjacency_edge(adjacency=adj, sort=True, copy_adjacency=True)
    return result


def construct_hiag(ref_genome, mut_genomes):
    ref_adjacencies = {}
    for adj in ref_genome.iter_adjacencies(adjacency_type=AdjacencyType.REFERENCE):
        aid = adj.stable_id_phased
        if aid not in ref_adjacencies:
            ref_adjacencies[aid] = adj
    nov_adjacencies = {}
    for mut_genome in mut_genomes:
        for adj in mut_genome.iter_adjacencies(adjacency_type=AdjacencyType.NOVEL):
            aid = adj.stable_id_phased
            if aid not in ref_adjacencies and aid not in nov_adjacencies:
                nov_adjacencies[aid] = adj
    segments = {}
    for s in ref_genome.iter_segments():
        sid = s.stable_id_hap
        if sid not in segments:
            segments[sid] = s
    result = HaplotypeSpecificIntervalAdjacencyGraph()
    for segment in segments.values():
        result.add_segment_edge(segment=segment, sort=True, copy_segment=True)
    for adj in ref_adjacencies.values():
        result.add_adjacency_edge(adjacency=adj, sort=True, copy_adjacency=True)
    for adj in nov_adjacencies.values():
        result.add_adjacency_edge(adjacency=adj, sort=True, copy_adjacency=True)
    return result


def construct_hiag_inflate_from_haploid_data(hapl_segments, hapl_adjacencies):
    dip_ref_adjacencies = []
    for adj in filter(lambda a: a.adjacency_type == AdjacencyType.REFERENCE, hapl_adjacencies):
        for ph in [Phasing.AA, Phasing.BB]:
            dip_adjacency = adj.get_non_phased_copy()
            dip_adjacency.extra[PHASING] = ph
            dip_ref_adjacencies.append(dip_adjacency)
    dip_nov_adjacencies = []
    for adj in filter(lambda a: a.adjacency_type == AdjacencyType.NOVEL, hapl_adjacencies):
        for ph in [Phasing.AA, Phasing.AB, Phasing.BA, Phasing.BB]:
            dip_adjacency = adj.get_non_phased_copy()
            dip_adjacency.extra[PHASING] = ph
            dip_nov_adjacencies.append(dip_adjacency)
    dip_segments = []
    for segment in hapl_segments:
        for h in [Haplotype.A, Haplotype.B]:
            dip_segment = segment.get_non_hap_copy()
            dip_segment.extra[HAPLOTYPE] = h
            dip_segments.append(dip_segment)
    for s in dip_segments:
        propagate_haplotype_segment_to_positions(segment=s, inplace=True)
    all_dip_adjacencies = dip_ref_adjacencies + dip_nov_adjacencies
    for a in all_dip_adjacencies:
        propagate_phasing_adjacency_to_positions(adjacency=a, inplace=True)
    hiag = HaplotypeSpecificIntervalAdjacencyGraph(segments=dip_segments, adjacencies=dip_ref_adjacencies + dip_nov_adjacencies)
    hiag.build_graph()
    return hiag


construct_hsiag = construct_hiag

HSIAG = HaplotypeSpecificIntervalAdjacencyGraph
HIAG = HSIAG



import argparse
import itertools
import os
import sys
from collections import defaultdict

current_file_level = 3
current_dir = os.path.dirname(os.path.realpath(__file__))
for _ in range(current_file_level):
    current_dir = os.path.dirname(current_dir)
sys.path.append(current_dir)

import rck
from rck.core.io import read_adjacencies_from_source, write_adjacencies_to_destination, EXTERNAL_NA_ID, stream_adjacencies_from_source, get_logging_cli_parser, \
    get_standard_logger_from_args
from rck.utils.adj.process import get_shared_nas_parser, Merger, iter_over_string_entries_from_source, get_extra_field_regexes, \
    filter_adjacencies_by_extra, \
    KEEP, REMOVE, refined_adjacencies_reciprocal
from rck.utils.adj.convert import get_chrs_regions_string_lists_from_source, get_chrs_regions_string_list_from_file, parse_segment_chr_region
from rck.utils.adj.process import filter_adjacencies_by_chromosomal_regions, filter_adjacencies_by_size


def main():
    parser = argparse.ArgumentParser(prog="RCK-UTILS-ADJ-process")
    parser.add_argument('--version', action='version', version=rck.version)
    ####
    shared_parser = get_shared_nas_parser()
    cli_logging_parser = get_logging_cli_parser()
    shared_parser.add_argument("--output", "-o", dest="rck_adj_file", type=argparse.FileType("wt"), default=sys.stdout)
    shared_parser.add_argument("--no-sort", action="store_false", dest="sort")
    ####
    subparsers = parser.add_subparsers(title="commands", dest="command")
    subparsers.required = True
    ####
    filter_parser = subparsers.add_parser("filter", parents=[shared_parser, cli_logging_parser])
    filter_parser.add_argument("rck_adj", type=argparse.FileType("rt"), nargs="+", default=[sys.stdin])
    filter_parser.add_argument("--keep-extra-field-regex", nargs="+", default=None)
    filter_parser.add_argument("--keep-extra-field-regex-file", type=argparse.FileType, default=None)
    filter_parser.add_argument("--keep-extra-field-missing-strategy", choices=[KEEP, REMOVE], default=KEEP)
    filter_parser.add_argument("--remove-extra-field-regex", nargs="+", default=None)
    filter_parser.add_argument("--remove-extra-field-regex-file", type=argparse.FileType, default=None)
    filter_parser.add_argument("--remove-extra-field-missing-strategy", choices=[KEEP, REMOVE], default=KEEP)
    filter_parser.add_argument("--min-size", type=int, default=0)
    filter_parser.add_argument("--max-size", type=int, default=1000000000)
    filter_parser.add_argument("--no-allow-inter-chr", action="store_false", dest="allow_inter_chr")
    filter_parser.add_argument("--size-extra-field")
    filter_parser.add_argument("--size-extra-field-no-abs", action="store_false", dest="size_extra_field_abs")
    filter_parser.add_argument("--size-extra-seq-field")
    # filter_parser.add_argument("--merged-field", default="origin_ids")
    # filter_parser.add_argument("--merged-distinct-origin-min-cnt", type=int, default=-1)
    # filter_parser.add_argument("--merged-distinct-origin-min-cnt-missing-strategy", choices=[KEEP, REMOVE], default=KEEP)
    # filter_parser.add_argument("--merged-origin-regex", default=".*_(?P<source>.*)")
    ####
    cat_parser = subparsers.add_parser("cat", parents=[shared_parser, cli_logging_parser], help="Concatenate Adjacencies in input files (NOTE: different from \"merge\")")
    cat_parser.add_argument("rck_adj", type=argparse.FileType("rt"), nargs="+", default=[sys.stdin])
    cat_parser.add_argument("--enforce-unique-ids", action="store_true", dest="enforce_unique_ids")
    cat_parser.add_argument("--id-collision-strategy", choices=['skip', 'error'], default='error')
    ####
    reciprocal_parser = subparsers.add_parser("reciprocal", parents=[shared_parser, cli_logging_parser], help="ensure that reciprocal novel adjacencies are treated as such")
    reciprocal_parser.add_argument("rck_adj", type=argparse.FileType("rt"), default=sys.stdin)
    reciprocal_parser.add_argument("--max-distance", type=int, default=50)
    ####
    # merge = subparsers.add_parser("merge", parents=[shared_parser, cli_logging_parser], help="EXPERIMENTAL! Merge Adjacencies in input files (NOTE: different from \"cat\").")
    # merge.add_argument("rck_adj", type=argparse.FileType("rt"), nargs="+", default=[sys.stdin])
    # merge.add_argument("--max-distance", type=int, default=500)
    # merge.add_argument("--enforce-sv-types", action="store_true", dest="enforce_sv_types", usage=argparse.SUPPRESS)
    # merge.add_argument("--merged-field", default="origin_ids")
    # merge.add_argument("--merged-id-template", default="{cnt}_merged")
    ####
    # test_merge = subparsers.add_parser("test-merge", parents=[shared_parser, cli_logging_parser], usage=argparse.SUPPRESS)
    # test_merge.add_argument("sources", type=argparse.FileType("rt"), nargs="+")
    # test_merge.add_argument("--merged", type=argparse.FileType("rt"), required=True)
    # test_merge.add_argument("--max-distance", default=500)
    # test_merge.add_argument("--merged-field", default="origin_ids")
    # test_merge.add_argument("--origin-sep", default=",")
    # test_merge.add_argument("--origin-regex", default=".*_(?P<source>.*)")
    args = parser.parse_args()
    logger = get_standard_logger_from_args(args=args, program_name="RCK-UTILS-ADK-process")
    processed_adjacencies = []
    if args.o_extra_fields is None or len(args.o_extra_fields) == 0 or args.o_extra_fields == ",":
        extra = None
    elif args.o_extra_fields != "all":
        extra = args.o_extra_fields.split(",")
    else:
        extra = args.o_extra_fields
    if args.command == "merge":
        all_adjacencies = []
        for rck_adj_source in args.rck_adj:
            all_adjacencies.extend(read_adjacencies_from_source(source=rck_adj_source))
        merger = Merger(origin_ids_field=args.merged_field)
        for na in all_adjacencies:
            merger.add_adjacency(adjacency=na, max_distance=args.max_distance)
        processed_adjacencies = merger.get_merged_adjacencies(merged_template=args.merged_id_template)
    elif args.command == "test-merge":
        source_adjacencies = []
        for rck_adj_source in args.sources:
            source_adjacencies.extend(read_adjacencies_from_source(source=rck_adj_source))
        merged_nas = read_adjacencies_from_source(source=args.merged)
        source_nas_by_ids = {adj.extra.get(EXTERNAL_NA_ID, adj.idx): adj for adj in source_adjacencies}
        logger.info("Total number of merged nas :: {mlen}".format(mlen=len(merged_nas)))
        logger.info("Total number of source nas :: {slen}".format(slen=len(source_adjacencies)))
        source_nas_groups = defaultdict(list)
        cnt = 0
        for merged_na in merged_nas:
            if args.merged_field not in merged_na.extra:
                logger.info("no source ids field {{{field}}} in merged_na {naid}".format(field=args.merged_field, naid=merged_na.extra.get(EXTERNAL_NA_ID, merged_na.idx)))
                continue
            source_ids = merged_na.extra[args.merged_field].split(args.origin_sep)
            for source_id in source_ids:
                source_nas_groups[source_id].append(merged_na.extra.get(EXTERNAL_NA_ID, merged_na.idx))
            for ona1_id, ona2_id in itertools.combinations(source_ids, 2):
                ona1 = source_nas_by_ids[ona1_id]
                ona2 = source_nas_by_ids[ona2_id]
                if not Merger.adjacencies_mergeable(adjacency1=ona1, adjacency2=ona2, max_distance=args.max_distance):
                    cnt += 1
        logger.info("Non mergable pairs cnt :: {cnt}".format(cnt=cnt))
        for source_id, groups in source_nas_groups.items():
            if len(groups) > 1:
                logger.info("source na id {source_id} is present in several merged adjacencies {merged_ids}".format(source_id=source_id, merged_ids=str(groups)))
    elif args.command == "cat":
        adjacencies = itertools.chain(*(stream_adjacencies_from_source(source=rck_adj_source) for rck_adj_source in args.rck_adj))
        if args.enforce_unique_ids:
            processed_ids = set()
            adjacencies = []
            for adj in adjacencies:
                aid = adj.extra.get(EXTERNAL_NA_ID, adj.idx)
                if aid in processed_ids:
                    logger.debug("Adjacency id {aid} has been encountered more than once".format(aid=aid))
                    if args.id_collision_strategy == "skip":
                        continue
                    elif args.id_collision_strategy == "error":
                        raise ValueError("More than one adjacency with id {aid}".format(aid=aid))
                adjacencies.append(adj)
                processed_ids.add(aid)
            adjacencies = adjacencies
        write_adjacencies_to_destination(destination=args.rck_adj_file, adjacencies=adjacencies, extra=extra, sort_adjacencies=args.sort)
        exit(0)
    elif args.command == "filter":
        logger.info("Filtering input adjacencies from following sources {sources}".format(sources=",".join(map(str, args.rck_adj))))
        adjacencies = itertools.chain(*(stream_adjacencies_from_source(source=rck_adj_source) for rck_adj_source in args.rck_adj))
        include_chrs_regions_strings = []
        exclude_chrs_regions_strings = []
        if args.chrs_include is not None:
            for chrs_lists in args.chrs_include:
                for chrs_list in chrs_lists:
                    for chr_name in chrs_list.split(","):
                        include_chrs_regions_strings.append(chr_name)
        if args.chrs_include_file is not None:
            for chr_name in get_chrs_regions_string_lists_from_source(source=args.chrs_include_file):
                include_chrs_regions_strings.append(chr_name)
        if args.chrs_exclude is not None:
            for chrs_lists in args.chrs_exclude:
                for chrs_list in chrs_lists:
                    for chr_name in chrs_list.split(","):
                        exclude_chrs_regions_strings.append(chr_name)
        if args.chrs_exclude_file is not None:
            for chr_name in get_chrs_regions_string_list_from_file(file_name=args.chrs_exclude_file):
                exclude_chrs_regions_strings.append(chr_name)
        include_regions = [parse_segment_chr_region(string) for string in include_chrs_regions_strings]
        exclude_regions = [parse_segment_chr_region(string) for string in exclude_chrs_regions_strings]
        adjacencies = filter_adjacencies_by_chromosomal_regions(adjacencies=adjacencies, include=include_regions, exclude=exclude_regions,
                                                                include_both=args.include_both, exclude_both=args.exclude_both)
        keep_extra_field_entries = args.keep_extra_field_regex if args.keep_extra_field_regex is not None else []
        if args.keep_extra_field_regex_file is not None:
            keep_extra_field_entries.extend(list(iter_over_string_entries_from_source(source=args.keep_extra_field_regex_file)))
        remove_extra_field_entries = args.remove_extra_field_regex if args.remove_extra_field_regex is not None else []
        if args.remove_extra_field_regex_file is not None:
            remove_extra_field_entries.extend(list(iter_over_string_entries_from_source(source=args.remove_extra_field_regex_file)))
        keep_extra_field = get_extra_field_regexes(string_entries=keep_extra_field_entries)
        remove_extra_field = get_extra_field_regexes(string_entries=remove_extra_field_entries)
        adjacencies = filter_adjacencies_by_extra(adjacencies=adjacencies,
                                                  keep_extra_field=keep_extra_field, keep_extra_field_missing_strategy=args.keep_extra_field_missing_strategy,
                                                  remove_extra_field=remove_extra_field, remove_extra_field_missing_strategy=args.remove_extra_field_missing_strategy)
        adjacencies = filter_adjacencies_by_size(adjacencies=adjacencies, min_size=args.min_size, max_size=args.max_size, size_extra_field=args.size_extra_field,
                                                 size_extra_seq_field=args.size_extra_seq_field, allow_inter_chr=args.allow_inter_chr,
                                                 size_extra_field_abs=args.size_extra_field_abs)
        write_adjacencies_to_destination(destination=args.rck_adj_file, adjacencies=adjacencies, sort_adjacencies=False, extra=extra)
        exit(0)
    elif args.command == "reciprocal":
        adjacencies = read_adjacencies_from_source(source=args.rck_adj)
        processed_adjacencies = refined_adjacencies_reciprocal(novel_adjacencies=adjacencies, max_distance=args.max_distance, inplace=True)

    if len(processed_adjacencies) > 0:
        write_adjacencies_to_destination(destination=args.rck_adj_file, adjacencies=processed_adjacencies, extra=extra, sort_adjacencies=args.sort)


if __name__ == "__main__":
    main()



import csv
import math

from rck.core.structures import SegmentCopyNumberProfile, Segment, Haplotype
from rck.utils.adj.convert import strip_chr

BATTENBERG_SAMPLE_NAME = "sample"
BATTENBERG_CHROMOSOME = "chr"
BATTENBERG_START_POSITION = "startpos"
BATTENBERG_END_POSITION = "endpos"
BATTENBERG_CLONE1_CN_A = "nMaj1_A"
BATTENBERG_CLONE1_CN_B = "nMin1_A"
BATTENBERG_CLONE2_CN_A = "nMaj2_A"
BATTENBERG_CLONE2_CN_B = "nMin2_A"


def battenberg_force_non_negativity(cn):
    return cn if cn >= 0 else 0


def battenberg_get_subclonal_cn(subclonal_cn_string, clonal_cn_int):
    if subclonal_cn_string == "NA":
        return clonal_cn_int
    return battenberg_force_non_negativity(int(subclonal_cn_string))


def battenberg_get_scnt_from_battenberg_file(file_name, sample_name, separator="\t", chr_strip=True):
    with open(file_name, "rt") as source:
        return get_scnt_from_battenberg_source(source=source, sample_name=sample_name, separator=separator, chr_strip=chr_strip)


def get_scnt_from_battenberg_source(source, sample_name, separator="\t", chr_strip=True):
    clone1_name = "1"
    clone2_name = "2"
    scnt = {clone1_name: SegmentCopyNumberProfile(), clone2_name: SegmentCopyNumberProfile()}
    segments = []
    reader = csv.DictReader(source, delimiter=separator)
    for row in reader:
        if BATTENBERG_SAMPLE_NAME in row and row[BATTENBERG_SAMPLE_NAME] != sample_name:
            continue
        start_coordinate = int(row[BATTENBERG_START_POSITION])
        end_coordinate = int(row[BATTENBERG_END_POSITION])
        chromosome = row[BATTENBERG_CHROMOSOME]
        if chr_strip:
            chromosome = strip_chr(chr_string=chromosome)
        segment = Segment.from_chromosome_coordinates(chromosome=chromosome, start=start_coordinate, end=end_coordinate)
        clone1_scnp = scnt[clone1_name]
        clone2_scnp = scnt[clone2_name]
        cn1a = battenberg_force_non_negativity(int(row[BATTENBERG_CLONE1_CN_A]))
        cn1b = battenberg_force_non_negativity(int(row[BATTENBERG_CLONE1_CN_B]))
        clone1_scnp.set_cn_record_for_segment(segment=segment, cn=cn1a, haplotype=Haplotype.A)
        clone1_scnp.set_cn_record_for_segment(segment=segment, cn=cn1b, haplotype=Haplotype.B)
        cn2a = battenberg_get_subclonal_cn(subclonal_cn_string=row[BATTENBERG_CLONE2_CN_A], clonal_cn_int=cn1a)
        cn2b = battenberg_get_subclonal_cn(subclonal_cn_string=row[BATTENBERG_CLONE2_CN_B], clonal_cn_int=cn1b)
        clone2_scnp.set_cn_record_for_segment(segment=segment, cn=cn2a, haplotype=Haplotype.A)
        clone2_scnp.set_cn_record_for_segment(segment=segment, cn=cn2b, haplotype=Haplotype.B)
        segments.append(segment)
    return segments, scnt


def hatchet_get_clone_ids_from_file(file_name, sample_name, separator="\t", min_usage=0.01):
    result = set()
    candidates = []
    with open(file_name, "rt") as source:
        for line_cnt, line in enumerate(source):
            line = line.strip()
            data = line.split(separator)
            clone_data = data[6:]
            if line_cnt == 0:
                total_clone_cnt = int(len(clone_data) / 2)
                candidates = [str(cnt) for cnt in range(1, total_clone_cnt + 1)]
            if line.startswith("#"):
                continue
            sample = data[3]
            if sample != sample_name:
                continue
            for candidate_clone_id, clone_usage_str in zip(candidates, clone_data[1::2]):
                clone_usage = float(clone_usage_str)
                if clone_usage < min_usage:
                    continue
                result.add(candidate_clone_id)
            if sorted(result) == candidates:
                return sorted(result)
    return sorted(result)


def get_scnt_from_hatchet_file(file_name, sample_name, separator="\t", clone_ids=None, min_usage=0.01, chr_strip=True):
    if clone_ids is None:
        clone_ids = hatchet_get_clone_ids_from_file(file_name=file_name, sample_name=sample_name, separator=separator, min_usage=min_usage)
    with open(file_name, "rt") as source:
        return get_scnt_from_hatchet_source(source=source, separator=separator, clone_ids=clone_ids, chr_strip=chr_strip)


def get_scnt_from_hatchet_source(source, clone_ids, separator="\t", chr_strip=True):
    scnt = {clone_id: SegmentCopyNumberProfile() for clone_id in clone_ids}
    segments = []
    clone_id_mappings = {}
    for line_cnt, line in enumerate(source):
        line = line.strip()
        data = line.split(separator)
        clone_data = data[6:]
        if line_cnt == 0:
            total_clone_cnt = int(len(clone_data) / 2)
            candidates = [str(cnt) for cnt in range(1, total_clone_cnt + 1)]
            for position_cnt, candidate in enumerate(candidates):
                if candidate in clone_ids:
                    clone_id_mappings[candidate] = position_cnt
        clone_cn_strs = clone_data[::2]
        if line.startswith("#") or len(line) == 0:
            continue
        sample_name = data[3]
        if sample_name != sample_name:
            continue
        chromosome = data[0]
        if chr_strip:
            chromosome = strip_chr(chr_string=chromosome)
        start_coord = int(data[1])
        end_coord = int(data[2]) - 1
        segment = Segment.from_chromosome_coordinates(chromosome=chromosome, start=start_coord, end=end_coord)
        segments.append(segment)
        fid = segment.stable_id_non_hap
        for clone_id in clone_ids:
            cns_str = clone_cn_strs[clone_id_mappings[clone_id]]
            data = cns_str.split("|")
            cna = int(data[0])
            cnb = int(data[1])
            scnt[clone_id].set_cn_record(sid=fid, hap=Haplotype.A, cn=cna)
            scnt[clone_id].set_cn_record(sid=fid, hap=Haplotype.B, cn=cnb)
    return segments, scnt


REMIXT_CHROMOSOME = "chromosome"
REMIXT_START_POSITION = "start"
REMIXT_END_POSITION = "end"
REMIXT_CLONE1_CN_A = "major_1"
REMIXT_CLONE1_CN_B = "minor_1"
REMIXT_CLONE2_CN_A = "major_2"
REMIXT_CLONE2_CN_B = "minor_2"


def get_scnt_from_remixt_file(file_name, separator="\t", chr_strip=True):
    with open(file_name, "rt") as source:
        return get_scnt_from_remixt_source(source=source, separator=separator, chr_strip=chr_strip)


def get_scnt_from_remixt_source(source, separator="\t", chr_strip=True):
    segments = []
    clone1_id = "1"
    clone2_id = "2"
    scnt = {clone1_id: SegmentCopyNumberProfile(), clone2_id: SegmentCopyNumberProfile()}
    reader = csv.DictReader(source, delimiter=separator)
    for row in reader:
        chromosome = row[REMIXT_CHROMOSOME]
        if chr_strip:
            chromosome = strip_chr(chr_string=chromosome)
        start_coordinate = int(row[REMIXT_START_POSITION])
        end_coordinate = int(row[REMIXT_END_POSITION]) - 1
        segment = Segment.from_chromosome_coordinates(chromosome=chromosome, start=start_coordinate, end=end_coordinate)
        segments.append(segment)
        sid = segment.stable_id_non_hap
        clone_1_cn_a = int(row[REMIXT_CLONE1_CN_A])
        clone_1_cn_b = int(row[REMIXT_CLONE1_CN_B])
        clone_2_cn_a = int(row[REMIXT_CLONE2_CN_A])
        clone_2_cn_b = int(row[REMIXT_CLONE2_CN_A])
        clone1_scnp = scnt[clone1_id]
        clone2_scnp = scnt[clone2_id]
        clone1_scnp.set_cn_record(sid=sid, hap=Haplotype.A, cn=clone_1_cn_a)
        clone1_scnp.set_cn_record(sid=sid, hap=Haplotype.B, cn=clone_1_cn_b)
        clone2_scnp.set_cn_record(sid=sid, hap=Haplotype.A, cn=clone_2_cn_a)
        clone2_scnp.set_cn_record(sid=sid, hap=Haplotype.B, cn=clone_2_cn_b)
    return segments, scnt


TITAN_CHROMOSOME = "Chromosome"
TITAN_START_POSITION = "Start"
TITAN_END_POSITION = "End"
TITAN_MAJOR_CN = "MajorCN"
TITAN_MINOR_CN = "MinorCN"
TITAN_CLONE_ID = "Clonal_Cluster"
TITAN_CORRECTED_CN = "Corrected_Copy_Number"
TITAN_SAMPLE_NAME = "Sample"


def titan_get_clone_ids_from_file(file_name, sample_name, separator="\t"):
    with open(file_name, "rt") as source:
        result = set()
        reader = csv.DictReader(source, delimiter=separator)
        for row in reader:
            if row[TITAN_SAMPLE_NAME] != sample_name:
                continue
            clone_id = row[TITAN_CLONE_ID]
            if clone_id != "NA":
                result.add(clone_id)
        return sorted(result)


def get_scnt_from_titan_file(file_name, sample_name, clone_ids=None, separator="\t", corrected_cn_fix="None", chr_strip=True):
    if clone_ids is None:
        clone_ids = titan_get_clone_ids_from_file(file_name=file_name, sample_name=sample_name, separator=separator)
    with open(file_name, "rt") as source:
        return get_scnt_from_titan_source(source=source, sample_name=sample_name, clone_ids=clone_ids, separator=separator, corrected_cn_fix=corrected_cn_fix, chr_strip=chr_strip)


def get_scnt_from_titan_source(source, sample_name, clone_ids, separator="\t", corrected_cn_fix="None", chr_strip=True):
    scnt = {clone_id: SegmentCopyNumberProfile() for clone_id in clone_ids}
    segments = []
    reader = csv.DictReader(source, delimiter=separator)
    for row in reader:
        if row[TITAN_SAMPLE_NAME] != sample_name:
            continue
        chromosome = row[TITAN_CHROMOSOME]
        if chr_strip:
            chromosome = strip_chr(chr_string=chromosome)
        segment = Segment.from_chromosome_coordinates(chromosome=chromosome, start=int(row[TITAN_START_POSITION]), end=int(row[TITAN_END_POSITION]))
        sid = segment.stable_id_non_hap
        segments.append(segment)
        major_cn, minor_cn = int(row[TITAN_MAJOR_CN]), int(row[TITAN_MINOR_CN])
        if minor_cn > major_cn:
            minor_cn, major_cn = major_cn, minor_cn
        titan_clone_id = row[TITAN_CLONE_ID]
        corrected_cn = int(row[TITAN_CORRECTED_CN])
        for clone_id in clone_ids:
            scnp = scnt[clone_id]
            if titan_clone_id == clone_id:
                if major_cn + minor_cn != corrected_cn and corrected_cn_fix != "None":
                    diff = corrected_cn - major_cn - minor_cn
                    ###
                    # initialize as 0 when corrected_cn_fix strategy does not match any known, yet is not "None"
                    ###
                    major_cn_addition = 0
                    minor_cn_addition = 0
                    if corrected_cn_fix == "equal":
                        major_cn_addition = int(math.ceil(diff / 2))
                        minor_cn_addition = diff - major_cn_addition
                    elif corrected_cn_fix == "relative-dist":
                        relative_relation = minor_cn * 1.0 / major_cn
                        major_cn_addition = int(math.ceil(diff / (1 + relative_relation)))
                        minor_cn_addition = diff - major_cn_addition
                    major_cn += major_cn_addition
                    minor_cn += minor_cn_addition
                scnp.set_cn_record(sid=sid, hap=Haplotype.A, cn=major_cn)
                scnp.set_cn_record(sid=sid, hap=Haplotype.B, cn=minor_cn)
            else:
                scnp.set_cn_record(sid=sid, hap=Haplotype.A, cn=1)
                scnp.set_cn_record(sid=sid, hap=Haplotype.B, cn=1)
    return segments, scnt



import argparse

import os
import sys

current_file_level = 2
current_dir = os.path.dirname(os.path.realpath(__file__))
for _ in range(current_file_level):
    current_dir = os.path.dirname(current_dir)
sys.path.append(current_dir)

from rck.core.structures import aligned_scnts, cn_distance_inter_scnt
from rck.core.io import read_scnt_from_source


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--scnt1", type=argparse.FileType("rt"), required=True)
    parser.add_argument("--scnt2", type=argparse.FileType("rt"), required=True)
    parser.add_argument("--clone-ids")
    args = parser.parse_args()
    segments1, scnt1 = read_scnt_from_source(source=args.scnt1)
    segments2, scnt2 = read_scnt_from_source(source=args.scnt2)
    if args.clone_ids is None:
        clone_ids = sorted(set(scnt1.keys()) & set(scnt2.keys()))
    else:
        clone_ids = args.clone_ids.split(",")
    for clone_id in sorted(scnt2.keys()):
        scnt_clone_id = sorted(scnt1.keys())[0]
        ref_scnt2 = {scnt_clone_id: scnt2[clone_id]}
        scnt1 = {clone_id: scnt1[clone_id] for clone_id in clone_ids}
    # scnt2 = {clone_id: scnt2[clone_id] for clone_id in clone_ids}
    # segments_by_sample_names = {"sample1": segments1, "sample2": segments2}
    # scnts_by_sample_names = {"sample1": scnt1, "sample2": scnt2}
    # segments_by_sn, scnts_by_sn = aligned_scnts(segments_by_sample_names=segments_by_sample_names, scnts_by_sample_names=scnts_by_sample_names)
    # scnt1, scnt2 = scnts_by_sn["sample1"], scnts_by_sn["sample2"]
    # segments1, segments2 = segments_by_sn["sample1"], segments_by_sn["sample2"]
        clone_specific_cn_distance = cn_distance_inter_scnt(tensor1=scnt1, tensor2=ref_scnt2, segments=segments1)
        print(clone_id, clone_specific_cn_distance.values())


if __name__ == "__main__":
    main()
